/*
 * Avatar/0 graphics decoder
 *
 * Copyright (C) 2006, 2010 Magnus Bergman
 *
 * Author: Magnus Bergman <fabel@users.sourceforge.net>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */
/** @file
  Avatar/0 graphics decoder.
*/
#include <stdlib.h>                     /* malloc() */

#include "avt0_dec.h"
#include "term_if.h"

struct _Avatar0Decoder{
    const TerminalInterface* term_if;
    Terminal* term;

    int incode;
    char code[10];
};

Avatar0Decoder* avatar0_decoder_new(const TerminalInterface* term_if,Terminal* term){
    Avatar0Decoder* dec;
    dec = (Avatar0Decoder*)malloc(sizeof(Avatar0Decoder));
    dec->term_if = term_if;
    dec->term = term;
    dec->incode = 0;
    return dec;
}

int avatar0_decoder_write(Avatar0Decoder* dec,char ch){
    if(dec->incode || 12==ch || 22==ch || 25==ch){
        dec->code[dec->incode++] = ch;
        switch(dec->code[0]){
        case 12: /* clear screen (or current window) */
            dec->term_if->clrscr(dec->term);
            dec->incode = 0;
            return 1;
        case 22: /* extended code */
            if(dec->incode<2)
                return 1;
            switch(dec->code[1]){
            case 1: /* text attribute (color) */
                if(dec->incode<3)
                    return 1;
                dec->term_if->set_attribute(dec->term,dec->code[2]);
                dec->incode = 0;
                return 1;
            case 2: /* set blink */
                dec->term_if->background(dec->term,dec->term_if->get_background(dec->term)|0x08);
                dec->incode = 0;
                return 1;
            case 3: /* go up */
                dec->term_if->up(dec->term,1);
                dec->incode = 0;
                return 1;
            case 4: /* go down */
                dec->term_if->down(dec->term,1);
                dec->incode = 0;
                return 1;
            case 5: /* go left */
                dec->term_if->left(dec->term,1);
                dec->incode = 0;
                return 1;
            case 6: /* go right */
                dec->term_if->right(dec->term,1);
                dec->incode = 0;
                return 1;
            case 7: /* clear end of line */
                dec->term_if->clreol(dec->term);
                dec->incode = 0;
                return 1;
            case 8: /* gotoxy */
                if(dec->incode<4)
                    return 1;
                dec->term_if->gotoxy(dec->term,dec->code[3],dec->code[2]);
                dec->incode = 0;
                return 1;
            default: /* unknown code */
                /* screen_write('?'); */
                dec->incode = 0;
                return 0;
            }
        case 25: /* repeat char */
            if(dec->incode<3)
                return 1;
            while(dec->code[2]--)
                dec->term_if->write(dec->term,dec->code[1]);
            dec->incode = 0;
            return 1;
#if 0
        default: /* unknown code, how did this get in here */
            dec->term_if->write(dec->term,'!');
            return 1;
#endif
        }
        return 1;
    }
    return 0;
}
