#ifndef doc_sprite_h
#define doc_sprite_h

#include "frame_tags.h"
#include "layer.h"
#include "palette.h"
#include "pixel_format.h"

namespace doc{
    typedef std::vector<Palette*> PalettesList;
    typedef PalettesList::iterator PalettesIter;
    typedef PalettesList::const_iterator PalettesConstIter;

    class Sprite{
    private:
        PixelFormat _format;
        int _width;
        int _height;
        int _frame_count;
        int _colors;
        int _transparent_color;
        LayerFolder *_folder;
        PalettesList _palettes;
        FrameTags _frame_tags;
        int _default_duration;
        std::vector<int> _duration;
    public:
        Sprite(PixelFormat, int width, int height, int colors);

        PixelFormat pixelFormat() const {return _format; };

        int width() const { return _width; }
        int height() const { return _height; }

        int transparentColor() const { return _transparent_color; }
        void setTransparentColor(int c) { _transparent_color = c; }

        /* --- layers --- */
        LayerFolder* folder()const{ return _folder; }

        Layer *indexToLayer(int) const;
        int layerToIndex(const Layer*) const;

        /* --- palettes --- */
        Palette *palette(int frame) const;
        void setPalette(const Palette *pal, bool truncate);

        /* --- frames --- */
        int totalFrames() const { return _frame_count; }
        void setTotalFrames(int);
        int frameDuration(int i) const { return _duration[i]; }
        void setFrameDuration(int frame,int msecs);
        void setDurationForAllFrames(int msecs);

        const FrameTags &frameTags() const { return _frame_tags; }
        FrameTags &frameTags() { return _frame_tags; }
    };
}

#endif /* doc_sprite_h */

