/*
 * Copyright (C) 2019  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <stdlib.h>                     /* malloc(), free() */

#include "cons_if.h"
#include "debug.h"
#include "petscii_dec.h"


struct _PetsciiDecoder{
    const ConsoleInterface* cons_vt;
    Console* cons;
    union{
        PetsciiState state;
        unsigned int attr:CONSOLE_BITS_ATTR;
    };
    int x;
    int y;
};


PetsciiDecoder* petscii_decoder_new(const ConsoleInterface* cons_vt,Console* cons){
    PetsciiDecoder* dec = (PetsciiDecoder*)malloc(sizeof(PetsciiDecoder));
    dec->cons_vt = cons_vt;
    dec->cons = cons;
    dec->attr = 0;
    dec->x = 0;
    dec->y = 0;
    return dec;
}


void petscii_decoder_free(PetsciiDecoder* dec){
    DEBUG_p("sizeof PetsciiState: %d",(int)sizeof(PetsciiState));
    free(dec);
}


void petscii_decoder_write(PetsciiDecoder* dec,int ch){
    //DEBUG_p("ch = 0x%x",ch);
    switch(ch){
    case 5: /* white */
        dec->state.color = 1;
        return;
    case 14: /* text mode */
        DEBUG_p("text mode");
        dec->state.unshifted = 0;
        return;
    case 18: /* reverse on */
        dec->state.reverse = 1;
        return;
    case 28: /* red */
        dec->state.color = 2;
        return;
    case 30: /* green */
        dec->state.color = 5;
        return;
    case 31: /* blue */
        dec->state.color = 6;
        return;
    case 129: /* orange */
        dec->state.color = 8;
        return;
    case 142: /* graphics */
        DEBUG_p("graphics");
        dec->state.unshifted = 1;
        return;
    case 144: /* black */
        dec->state.color = 0;
        return;
    case 146: /* reverse off */
        dec->state.reverse = 0;
        return;
    case 147: /* clear */
        return;
    case 149: /* brown */
        dec->state.color = 9;
        return;
    case 150: /* light red */
        dec->state.color = 10;
        return;
    case 151: /* dark gray */
        dec->state.color = 11;
        return;
    case 152: /* middle gray */
        dec->state.color = 12;
        return;
    case 153: /* light green */
        dec->state.color = 13;
        return;
    case 154: /* light blue */
        dec->state.color = 14;
        return;
    case 155: /* light gray */
        dec->state.color = 15;
        return;
    case 156: /* purple */
        dec->state.color = 4;
        return;
    case 158: /* yellow */
        dec->state.color = 7;
        return;
    case 159: /* cyan */
        dec->state.color = 3;
        return;
    default:
        switch(ch>>4){
        case 2:
        case 3:
            ch += dec->state.unshifted ? 0x80 : 0;
            break;
        case 4:
        case 5:
            ch += dec->state.unshifted ? 0x40 : -0x40;
            break;
        case 6:
        case 7:
            ch += dec->state.unshifted ? 0x60 : -0x20;
            break;
        case 10:
        case 11:
            ch += dec->state.unshifted ? 0x40 : -0x40;
            break;
        default:
            DEBUG_p("unchandled code 0x%x (%d)",ch,ch);
            return;
        }
        DEBUG_p("attr: %08x, color: %d",dec->attr,dec->state.color);
        dec->cons_vt->charxya(dec->cons,dec->x,dec->y,dec->attr,ch);
        if(++dec->x >= 40){
            dec->x = 0;
            if(++dec->y >= 25)
                dec->y = 0;
        }
        break;
    }
}
