/*
 * Author: Magnus Bergman <magnus@snisurset.net>
 *
 * This file is public domain.
 */
#ifndef cgmtypes_h
#define cgmtypes_h

#include <loa-buffer.h>
#include <nile.h>

#define ACT4_GRADIENT_ACTION    1

typedef uint32_t color_t;

typedef enum{
    RT_FLOAT = 0,
    RT_FIXED = 1
}real_type_t;

typedef enum{
    VDC_INT = 0,
    VDC_REAL = 1
}vdc_type_t;

typedef enum { SM_ABSTRACT = 0, SM_METRIC = 1 } ScalingMode;
typedef enum { CSM_INDEXED = 0, CSM_DIRECT = 1 } ColorSelectionMode;
//LineWidthSpecificationMode ABS SCALED FRACTIONAL MM
//MarkerSizeSpecificationMode ABS SCALED FRACTIONAL MM
//EdgeWidthSpecificationMode ABS SCALED FRACTIONAL MM
typedef enum { DVPM_FRACTION = 0, DVPM_METRIC = 1, DVPM_DEVICE = 2 } DeviceViewPortMode;
typedef enum { DVPM_NOT_FORCED = 0, DVPM_FORCED = 1 } DeviceViewPortMap;
typedef enum { DVPMH_LEFT = 0, DVPMH_CENTER = 1, CVPMH_RIGHT = 2 } DeviceViewPortMapH;
typedef enum { DVPMV_BOTTOM = 0, DVPMV_CENTER = 1, DVPMV_TOP = 2 } DeviceViewPortMapV;
typedef enum { CI_OFF = 0, CI_ON = 1 } ClipIndicator;
typedef enum{
    /* Standard */
    LT_SOLID = 1,
    LT_DASH = 2,
    LT_DOT = 3,
    LT_DASHDOT = 4,
    LT_DASHDOTDOT = 5,
    /* GDSF Styles */
    LT_NONE = -4,
    LT_DOTDOTSPACE = -3,
    LT_LONGDASH = -2,
    LT_DASHDASHDOT = -1
}line_type_t;
typedef enum { MT_POINT = 1, MT_PLUS = 2, MT_STAR = 3, MT_CIRCLE = 4, MT_CROSS = 5 } MarkerType;
typedef enum { CCA_BASIC_7 = 0, CCA_BASIC_8 = 1, CCA_EXT_7 = 2, CCA_EXT_8 = 3 } CharacterCodingA;
typedef enum { TPR_STRING = 0, TPR_CHARACTER = 1, TPR_STROKE = 2, TPR_UNDEFINED = 0xffff } TextPrecision;
typedef enum { TAH_NORMAL = 0, TAH_LEFT = 1, TAH_CENTER = 2, TAH_RIGHT = 3, TAH_CONT = 4 } TextAlignmentH;
typedef enum { TAV_NORMAL = 0, TAV_TOP = 1, TAV_CAP = 2, TAV_HALF = 3, TAV_BASE = 4, TAV_BOTTOM = 5, TAV_CONT = 6 } TextAlignmentV;
typedef enum{
    FIS_HOLLOW = 0,
    FIS_SOLID = 1,
    FIS_PATTERN = 2,
    FIS_HATCH = 3,
    FIS_EMPTY = 4,
    FIS_GEOPATTERN = 5,
    FIS_INTERPOLATED = 6,
    FIS_GRADIENT = 7
}interior_style_t;
typedef enum {
    HI_HORIZONTAL = 1,
    HI_VERTICAL = 2,
    HI_POSITIVE_SLOPE = 3,
    HI_NEGATIVE_SLOPE = 4,
    HI_HV_CROSS = 5,
    HI_PN_CROSS = 6,
}hatch_index_t;

typedef enum { CT_PIE = 0, CT_CHORD = 1, CT_xxx = 2 } close_type_t;
typedef enum { TPR_RIGHT = 0, TPR_LEFT = 1, TPR_UP = 2, TPR_DOWN = 3 } TextPath;
typedef enum { UM_OFF = 0, UM_LOW = 1, UM_HIGH = 2, UM_STRIKEOUT = 4, UM_OVERSCORE = 8 } UnderlineMode;
typedef enum { FF_NOT_FINAL = 0, FF_FINAL = 1 } FinalFlag;
typedef enum { SM_ABSOLUTE = 0, SM_SCALED = 1 } SpecMode;
typedef enum { LCT_BUTT = 0, LCT_ROUND = 1, LCT_SQUARE = 2, LCT_TRIANGLE = 3, LCT_ARROW = 4, LCT_NONE = -1 } LineCapType;
typedef enum { LJT_MITER = 0, LJT_ROUND = 1, LJT_BEVEL = 2, LJT_NONE = -1 } LineJoinType;
#if 0
typedef enum{
    /* Standard */
    ET_SOLID = 1,
    ET_DASH = 2,
    ET_DOT = 3,
    ET_DASHDOT = 4,
    ET_DASHDOTDOT = 5,
    /* GDSF Styles */
    ET_NONE = -4,
    ET_DOTDOTSPACE = -3,
    ET_LONGDASH = -2,
    ET_DASHDASHDOT = -1
}EdgeType;
#endif
typedef enum { EV_OFF = 0, EV_ON = 1 } EdgeVisibility;
#define EF_VISIBLE_BIT (1<<0)
#define EF_CLOSE_BIT (1<<1)
typedef enum { EF_INVISIBLE = 0, EF_VISIBLE = EF_VISIBLE_BIT, EF_CLOSE_VISIBLE = EF_VISIBLE_BIT|EF_CLOSE_BIT } edge_flag_t;
typedef enum { T_OFF = 0, T_ON = 1 } Transparency;

typedef enum { ASF_INDIVIDUAL = 0, ASF_BUNDLED = 1 } asf_value_t;

typedef struct{
    loa_point_t pos;
    loa_point_t size;
}rect_t;

typedef struct{
    loa_point_t *point;
    loa_segment_t *segment;
    int segment_count;
}path_t;

typedef struct{
    loa_point_t pos;
    loa_point_t size;
    double orientation;
    double startangle;
    double endangle;
    close_type_t close_type;
    int negative;
}arc_t;

struct _loa_object_t{
    enum{
        OT_RECT,
        OT_PATH,
        OT_ARC,
        OT_IMAGE
    }type;
    union{
        rect_t rect;
        path_t path;
        arc_t arc;
        struct{
            loa_point_t *p;
            nile_t ni;
        }image;
    };
};

#endif /* cgmtypes_h */
