/*
 * Copyright (C) 2021  Magnus Bergman <magnus@snisurset.net>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */
#include <math.h>

#include "loa-matrix.h"


void loa_matrix_set_identity(loa_matrix_t *m)
{
    m->xx = 1;
    m->yx = 0;
    m->xy = 0;
    m->yy = 1;
    m->x0 = 0;
    m->y0 = 0;
}


void loa_matrix_multiply(loa_matrix_t *m1, const loa_matrix_t *m2)
{
    loa_matrix_t m;
    m.xx = m1->xx * m2->xx + m1->yx * m2->xy;
    m.yx = m1->xx * m2->yx + m1->yx * m2->yy;
    m.xy = m1->xy * m2->xx + m1->yy * m2->xy;
    m.yy = m1->xy * m2->yx + m1->yy * m2->yy;
    m.x0 = m1->x0 * m2->xx + m1->y0 * m2->xy + m2->x0;
    m.y0 = m1->x0 * m2->yx + m1->y0 * m2->yy + m2->y0;
    *m1 = m;
}


void loa_matrix_translate(loa_matrix_t *m, double tx, double ty)
{
    m->x0 += tx;
    m->y0 += ty;
}


void loa_matrix_scale(loa_matrix_t *m, double sx, double sy)
{
    m->xx *= sx;
    m->yx *= sy;
    m->xy *= sx;
    m->yy *= sy;
    m->x0 *= sx;
    m->y0 *= sy;
}


void loa_matrix_rotate(loa_matrix_t *m, double a)
{
    double c = cos(a);
    double s = sin(a);
    loa_matrix_t m2;
    m2.xx = c;
    m2.yx = -s;
    m2.xy = s;
    m2.yy = c;
    m2.x0 = 0;
    m2.y0 = 0;
    loa_matrix_multiply(m, &m2);
}


double loa_matrix_apply_x(const loa_matrix_t *m, double x)
{
    return m->xx * x + m->x0;
}


double loa_matrix_apply_y(const loa_matrix_t *m, double y)
{
    return m->yy * y + m->y0;
}


void loa_matrix_apply_point(const loa_matrix_t *m, loa_point_t *p)
{
    double x = p->x;
    double y = p->y;
    p->x = m->xx * x + m->xy * y + m->x0;
    p->y = m->yx * x + m->yy * y + m->y0;
}
