/*
 * Copyright (C) 2021  Magnus Bergman <magnus@snisurset.net>
 * Copyright (C) 2017  Bastien Nocera <hadess@hadess.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */
#ifndef abydos_archive_h
#define abydos_archive_h

#ifdef USE_LIBARCHIVE
# include <archive.h>
# include <archive_entry.h>
#endif
#ifdef USE_LIBUNARR
# include <unarr.h>
#endif

#ifdef USE_ACEUNPACK
# include "aceunpack.h"
#endif
#include "dmc_unrar.h"

typedef enum {
#if defined(USE_LIBARCHIVE) || defined(USE_LIBUNARR)
    AT_7Z,
#endif
#ifdef USE_ACEUNPACK
    AT_ACE,
#endif
    AT_RAR,
#if defined(USE_LIBARCHIVE) || defined(USE_LIBUNARR)
    AT_TAR,
    AT_ZIP
#endif
}abydos_archive_type_t;

typedef struct{
    abydos_archive_type_t type;
    union{
        struct{
            /* libarchive */
            struct archive *archive;
            struct archive_entry *entry;
        }libar;
#ifdef USE_ACEUNPACK
        aceunpack_t *ace;
#endif
#ifdef USE_LIBUNARR
        struct{
            /* unarr */
            ar_stream *stream;
            ar_archive *archive;
        }unarr;
#endif
        struct{
            dmc_unrar_archive archive;
            dmc_unrar_file_block *entry;
            char filename[4096];
            char *buffer;
            size_t buffer_size;
            off_t offset;
        }dmc;
    };
}abydos_archive_t;

abydos_archive_t *abydos_archive_init(abydos_archive_t*, abydos_archive_type_t, const char *filename);
void              abydos_archive_done(abydos_archive_t*);
int               abydos_archive_next(abydos_archive_t*);
const char       *abydos_archive_entry_name(abydos_archive_t*);
ssize_t           abydos_archive_entry_size(abydos_archive_t*);
int               abydos_archive_entry_is_encrypted(abydos_archive_t*);
ssize_t           abydos_archive_entry_read(abydos_archive_t*,void *buf, size_t bytes);

#endif /* __EV_ARCHIVE_H__ */
