/*
 * Copyright (C) 2021  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
extern "C"{
#define new new_
#include <abydos-plugin.h>
#undef new
#include <debug.h>
}

#include "decode.h"


struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    cairo_surface_t *surface;
};


static cairo_surface_t *
_surface_from_image(const Image &img)
{
    cairo_surface_t *surface = cairo_image_surface_create(CAIRO_FORMAT_ARGB32, img.w, img.h);
    int offs = 0;
    uint32_t *dst = (uint32_t*)cairo_image_surface_get_data(surface);
    int dst_skip = cairo_image_surface_get_stride(surface) / sizeof(uint32_t) - img.w;
    int y;
    DEBUG_p("channels: %d", (int)img.channel.size());
    for(y=0;y<img.h;++y){
        int x;
        for(x=0;x<img.w;++x){
            int a = img.channel[3].value_nocheck(y,x);
            int r = a * img.channel[2].data[offs] / 255;
            int g = a * img.channel[1].data[offs] / 255;
            int b = a * img.channel[0].data[offs] / 255;
            *dst++ = r << 0 | g << 8 | b << 16 | a << 24;
            ++offs;
        }
        dst += dst_skip;
    }
    cairo_surface_mark_dirty(surface);
    return surface;
}


extern "C" const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = [](){
            static const char *mime_types[] = {
                "image/x-fuif",
                NULL
            };
            return mime_types;
        },
        .new_ = [](const char *mime_type, abydos_plugin_info_t *info){
            auto *h = new abydos_plugin_handle_t;
            info->threadsafe = 1;
            h->info = info;
            h->surface = NULL;
            return h;
        },
        .free = [](abydos_plugin_handle_t *h){
            if(h->surface)
                cairo_surface_destroy(h->surface);
            delete h;
        },
        .create_from_data = [](abydos_plugin_handle_t *h, const char *data, size_t len){
            BlobReader io((uint8_t*)data, len);
            Image img;
            if(!fuif_decode(io, img))
                return -1;
            img.undo_transforms();
            h->info->width = img.w;
            h->info->height = img.h;
            h->surface = _surface_from_image(img);
            return 0;
        },
        .get_image_surface = [](abydos_plugin_handle_t *h,
                       int page,
                       cairo_bool_t *layer_visibility,
                       int variant,
                       double time,
                       int frame)
        {
            return cairo_surface_reference(h->surface);
        }
    };
    return &p;
}
