/*
 * Author: Magnus Bergman <magnus@snisurset.net>
 *
 * This file is public domain.
 */
#ifndef ni_xng_h
#define ni_xng_h

#include <endian.h>
#include <stddef.h>
#include <stdint.h>

#define JNG_SIGNATURE "\213JNG\r\n\032\n"
#define MNG_SIGNATURE "\212MNG\r\n\032\n"
#define PNG_SIGNATURE "\211PNG\r\n\032\n"

#if __BYTE_ORDER == __BIG_ENDIAN
# define XNG_ID(a,b,c,d) ((a<<24)|(b<<16)|(c<<8)|(d<<0))
#else
# define XNG_ID(a,b,c,d) ((a<<0)|(b<<8)|(c<<16)|(d<<24))
#endif

typedef int(*ni_xng_parse_func_t)(void *userdata, uint32_t id, const char *, size_t);

int ni_xng_parse(const char *signature, uint32_t endid, const char*, size_t len, int validate, ni_xng_parse_func_t, void *userdata);
#define ni_xng_parse_chunks(a,b,c,d,e) ni_xng_parse(NULL,0,a,b,c,d,e)
#define ni_jng_parse(a,b,c,d,e) ni_xng_parse(JNG_SIGNATURE,XNG_ID('I','E','N','D'),a,b,c,d,e)
#define ni_mng_parse(a,b,c,d,e) ni_xng_parse(MNG_SIGNATURE,XNG_ID('M','E','N','D'),a,b,c,d,e)
#define ni_png_parse(a,b,c,d,e) ni_xng_parse(PNG_SIGNATURE,XNG_ID('I','E','N','D'),a,b,c,d,e)

#endif /* ni_xng_h */
