/*
 * Copyright (C) 2020  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <debug.h>

#include "cebra.h"

/*

0x30
  0x02
  0x02
  0x30
    0x30
      0x02
      0x02
      0x02
      0x02
      0x30
        0x30
          0x30 (subpage)
            0x30 (subdateData)
              0x02
              0x04
              0x02
              0x30 (rowData)
                0x30
                  0x02 row number
                  0x04 row characters
                0x30
                  0x02
                  0x04 row characters
            0x30 (subpageFooter)
              0x02
              0x02
              0x02

*/


int
cebra_parse(int level, const uint8_t *data, size_t len, cebra_func_t userfunc, void *userdata)
{
    const uint8_t *end = data+len;
    for(;;){
        int type;
        int res;
        if(data+2 > end)
            return 0;
        type = data[0];
        switch(data[1]){
        case 0x81:
            if(data+3 > end)
                return -1;
            len = data[2];
            data += 3;
            break;
        case 0x82:
            if(data+4 > end)
                return -1;
            len = data[2]<<8|data[3];
            data += 4;
            break;
        case 0x83:
            if(data+5 > end)
                return -1;
            len = data[2]<<16|data[3]<<8|data[4];
            data += 5;
            break;
        case 0x84:
            if(data+6 > end)
                return -1;
            len = data[2]<<24|data[3]<<16|data[4]<<8|data[5];
            data += 6;
            break;
        default:
            len = data[1];
            data += 2;
            break;
        }
        if(data+len > end)
            return -1;
        DEBUG_p("level %d, block type: %d",level,type);
        if(48==type){
            res = cebra_parse(level+1, data, len, userfunc, userdata);
        }else{
            res = userfunc(userdata, level, type, data, len);
        }
        if(res)
            return res;
        data += len;
    }
}
