#ifndef gs_internal_h
#define gs_internal_h

#include "gs-types.h"

#include "gsccolor.h"
#include "gscms.h"
#include "gscompt.h"
#include "gscspace.h"
#include "gsdcolor.h"                   /* gx_device_color */
#include "gsgstate.h"
#include "gsimage.h"                    /* gx_image_enum_common_t */
#include "gspenum.h"                    /* gx_path */
#include "gsropt.h"                     /* gs_logical_operation_t */
#include "gstext.h"
#include "gstparam.h"                   /* gs_transparency_group_params_t */
#include "gsxfont.h"                    /* gx_xfont_procs */
#include "gxcindex.h"
#include "gxcmap.h"                     /* gx_cmapper_t */
#include "gxcvalue.h"                   /* gx_color_value */
#include "gxdda.h"                      /* gx_dda_fixed */
#include "gxpath.h"                     /* gx_fill_params */

#define GS_CHECK_VERSION(major,minor,micro) \
    (GHOSTSCRIPT_MAJOR_VERSION > (major) || \
    (GHOSTSCRIPT_MAJOR_VERSION == (major) && \
     GHOSTSCRIPT_MINOR_VERSION > (minor)) || \
    (GHOSTSCRIPT_MAJOR_VERSION == (major) && \
     GHOSTSCRIPT_MINOR_VERSION == (minor) && \
     GHOSTSCRIPT_MICRO_VERSION >= (micro)))

typedef gx_device_color gx_drawing_color;

/* Define a type for telling get_alpha_bits what kind of object */
/* is being rendered. */
typedef enum {
    go_text,
    go_graphics
} graphics_object_type;

/* Define an edge of a trapezoid.  Requirement: end.y >= start.y. */
typedef struct gs_fixed_edge_s {
    gs_fixed_point start;
    gs_fixed_point end;
} gs_fixed_edge;

/* Define the parameters passed to get_bits_rectangle. */
typedef struct gs_get_bits_params_s gs_get_bits_params_t;

/* copied from base/gxdevcli.h */
/* Define the structure for device color capabilities. */
typedef struct gx_device_anti_alias_info_s {
    int text_bits;              /* 1,2,4 */
    int graphics_bits;          /* ditto */
} gx_device_anti_alias_info;

typedef int32_t frac31; /* A fraction value in [-1,1].
    Represents a color (in [0,1])
    or a color difference (in [-1,1]) in shadings. */

/* copied from base/gxdevcli.h */
typedef enum {
    GX_CINFO_UNKNOWN_SEP_LIN = -1,
    GX_CINFO_SEP_LIN_NONE = 0,
    GX_CINFO_SEP_LIN = 1,
    GX_CINFO_SEP_LIN_NON_STANDARD = 2,
    GX_CINFO_SEP_LIN_STANDARD = 3
} gx_color_enc_sep_lin_t;

/* copied from base/gxdevcli.h */
#if GS_CHECK_VERSION(9,56,0)
typedef enum {
    GX_CINFO_OPMSUPPORTED_UNKNOWN = -1,
    GX_CINFO_OPMSUPPORTED_NOT = 0,
    GX_CINFO_OPMSUPPORTED = 1
} gx_cm_opmsupported_t;
#else
typedef enum {
    GX_CINFO_OPMODE_UNKNOWN = -1,
    GX_CINFO_OPMODE_NOT = 0,
    GX_CINFO_OPMODE = 1
} gx_cm_opmode_t;
#endif

/* component index value used to indicate no color component.  */
#define GX_CINFO_COMP_NO_INDEX 0xff

/* copied from base/gxdevcli.h */
typedef struct gx_device_color_info_s {

    /*
     * max_components is the maximum number of components for all
     * color models supported by this device. This does not include
     * any alpha components.
     */
    uchar max_components;

    /*
     * The number of color components. This does not include any
     * alpha-channel information, which may be integrated into
     * the gx_color_index but is otherwise passed as a separate
     * component.
     */
    uchar num_components;

    /*
     * Polarity of the components of the color space, either
     * additive or subtractive. This is used to interpret transfer
     * functions and halftone threshold arrays. Possible values
     * are GX_CM_POLARITY_ADDITIVE or GX_CM_POLARITY_SUBTRACTIVE
     */
    gx_color_polarity_t polarity;

    /*
     * The number of bits of gx_color_index actually used.
     * This must be <= ARCH_SIZEOF_COLOR_INDEX, which is usually 64.
     * Note that we now have planar devices which can support much more.
     * Changing this to a ushort to reflect this.
     */
    ushort depth;

    /*
     * Index of the gray color component, if any. The max_gray and
     * dither_gray values apply to this component only; all other
     * components use the max_color and dither_color values.
     *
     * Note:  This field refers to a 'gray' colorant because of the
     * past use of the max_gray/color and dither_grays/colors fields.
     * Prior to 8.00, the 'gray' values were used for monochrome
     * devices and the 'color' values for RGB and CMYK devices.
     * Ideally we would like to have the flexibiiity of allowing
     * different numbers of intensity levels for each colorant.
     * However this is not compatible with the pre 8.00 devices.
     * With post 8.00 devices, we can have two different numbers of
     * intensity levels.  For one colorant (which is specified by
     * the gray_index) we will use the max_gray/dither_grays values.
     * The remaining colorants will use the max_color/dither_colors
     * values.  The colorant which is specified by the gray_index
     * value does not have to be gray or black.  For example if we
     * have an RGB device and we want 32 intensity levels for red and
     * blue and 64 levels for green, then we can set gray_index to
     * 1 (the green colorant), set max_gray to 63 and dither_grays to
     * 64, and set max_color to 31 and dither_colors to 32.
     *
     * This will be GX_CINFO_COMP_NO_INDEX if there is no 'gray'
     * component.
     */
    byte gray_index;

    /*
     * max_gray and max_color are the number of distinct native
     * intensity levels, less 1, for the 'gray' and all other color
     * components, respectively. For nearly all current devices
     * that support both 'gray' and non-'gray' components, the two
     * parameters have the same value.  (See comment for gray_index.)
     *
     * dither_grays and dither_colors are the number of intensity
     * levels between which halftoning can occur, for the 'gra'y and
     * all other color components, respectively. This is
     * essentially redundant information: in all reasonable cases,
     * dither_grays = max_gray + 1 and dither_colors = max_color + 1.
     * These parameters are, however, extensively used in the
     * current code, and thus have been retained.
     *
     * Note that the non-'gray' values may now be relevant even if
     * num_components == 1. This simplifies the handling of devices
     * with configurable color models which may be set for a single
     * non-'gray' color model.
     */
    uint max_gray;              /* # of distinct color levels -1 */
    uint max_color;

    uint dither_grays;
    uint dither_colors;

    /*
     * Information to control super-sampling of objects to support
     * anti-aliasing.
     */
    gx_device_anti_alias_info anti_alias;

    /*
     * Flag to indicate if gx_color_index for this device may be divided
     * into individual fields for each component. This is almost always
     * the case for printers, and is the case for most modern displays
     * as well. When this is the case, halftoning may be performed
     * separately for each component, which greatly simplifies processing
     * when the number of color components is large.
     *
     * If the gx_color_index is separable in this manner, the comp_shift
     * array provides the location of the low-order bit for each
     * component. This may be filled in by the client, but need not be.
     * If it is not provided, it will be calculated based on the values
     * in the max_gray and max_color fields as follows:
     *
     *     comp_shift[num_components - 1] = 0,
     *     comp_shift[i] = comp_shift[i + 1]
     *                      + ( i == gray_index ? ceil(log2(max_gray + 1))
     *                                          : ceil(log2(max_color + 1)) )
     *
     * The comp_mask and comp_bits fields should be left empty by the client.
     * They will be filled in during initialization using the following
     * mechanism:
     *
     *     comp_bits[i] = ( i == gray_index ? ceil(log2(max_gray + 1))
     *                                      : ceil(log2(max_color + 1)) )
     *
     *     comp_mask[i] = (((gx_color_index)1 << comp_bits[i]) - 1)
     *                       << comp_shift[i]
     *
     * (For current devices, it is almost always the case that
     * max_gray == max_color, if the color model contains both gray and
     * non-gray components.)
     *
     * If separable_and_linear is not set, the data in the other fields
     * is unpredictable and should be ignored.
     */
    gx_color_enc_sep_lin_t separable_and_linear;
    byte                   comp_shift[GX_DEVICE_COLOR_MAX_COMPONENTS];
    byte                   comp_bits[GX_DEVICE_COLOR_MAX_COMPONENTS];
    gx_color_index         comp_mask[GX_DEVICE_COLOR_MAX_COMPONENTS];

    /*
     * Pointer to name for the process color model.
     */
    const char * cm_name;

    /*
     * Indicate if overprint mode is supported. This is only supported
     * for color models that have "DeviceCMYK" like behaivor: they support
     * the cyan, magenta, yellow, and black color components, and map the
     * components of a DeviceCMYK color space directly to these compoents.
     * Most such color spaces will have the name DeviceCMYK, but it is
     * also possible for DeviceN color models this behavior.
     *
     * If opmsupported has the value GX_CINFO_OPMSUPPORTED, the process_comps will
     * be a bit mask, with the (1 << i) bit set if i'th component is the
     * cyan, magenta, yellow, or black component and black_component will
     * be set to the index of a black component.
     */
#if GS_CHECK_VERSION(9,56,0)
    gx_cm_opmsupported_t opmsupported;
#else
    /*
     * A new mode GX_CINFO_OPMODE_RGB was added so that an RGB device could
     * simulate CMYK overprinting.
     */
    gx_cm_opmode_t opmode;
#endif
    gx_color_index process_comps;
    uint black_component;
    bool use_antidropout_downscaler;
} gx_device_color_info;


/* Structure for device procedures. */
//typedef struct gx_device_procs_s gx_device_procs;

/* Structure for page device procedures. */
/* Note that these take the graphics state as a parameter. */
typedef struct gx_page_device_procs_s {

#define dev_page_proc_install(proc)\
  int proc(gx_device *dev, gs_gstate *pgs)
    dev_page_proc_install((*install));

#define dev_page_proc_begin_page(proc)\
  int proc(gx_device *dev, gs_gstate *pgs)
    dev_page_proc_begin_page((*begin_page));

#define dev_page_proc_end_page(proc)\
  int proc(gx_device *dev, int reason, gs_gstate *pgs)
    dev_page_proc_end_page((*end_page));

} gx_page_device_procs;

/* Default procedures */
dev_page_proc_install(gx_default_install);
dev_page_proc_begin_page(gx_default_begin_page);
dev_page_proc_end_page(gx_default_end_page);

/* ----------- A stroked gradient recognizer data ----------*/

/* This structure is associated with a device for
   internal needs of the graphics library.
   The main purpose is to suppress stroke adjustment
   when painting a gradient as a set of parallel strokes.
   Such gradients still come from some obsolete 3d party software.
   See bug 687974,
 */

typedef struct gx_stroked_gradient_recognizer_s {
    bool stroke_stored;
    gs_fixed_point orig[4], adjusted[4]; /* from, to, width, vector. */
} gx_stroked_gradient_recognizer_t;

/* ---------------- Device structure ---------------- */

/*
 * Define the generic device structure.
 *
 * All devices start off with no device procs. The initialize_device_procs
 * function is called as soon as the device is copied from its prototype,
 * and that fills in the procs table with the device procedure pointers.
 * This can never fail.
 *
 * Next, the 'initialize_device' proc (if there is one) is called to do
 * the minimal initialization required for a device.
 *
 * The choice of the name Margins (rather than, say, HWOffset), and the
 * specification in terms of a default device resolution rather than
 * 1/72" units, are due to Adobe.
 *
 * ****** NOTE: If you define any subclasses of gx_device, you *must* define
 * ****** the finalization procedure as gx_device_finalize.  Finalization
 * ****** procedures are not automatically inherited.
 */
typedef struct gx_device_cached_colors_s {
    gx_color_index black, white;
} gx_device_cached_colors_t;

/*
 * Define the parameters controlling banding.
 */
/* if you make any additions/changes to this structure you need to make
   the appropriate additions/changes to gdev_space_params_cmp() */
typedef struct gx_band_params_s {
    int BandWidth;		/* (optional) band width in pixels */
    int BandHeight;		/* (optional) */
    size_t BandBufferSpace;	/* (optional) */
    size_t tile_cache_size;	/* (optional) */
} gx_band_params_t;

typedef enum {
    BandingAuto = 0,
    BandingAlways,
    BandingNever
} gdev_banding_type;

/* if you make any additions/changes to this structure you need to make
   the appropriate additions/changes to the gdev_space_params_cmp() */
typedef struct gdev_space_params_s {
    size_t MaxBitmap;           /* max size of non-buffered bitmap */
    size_t BufferSpace;         /* space to use for buffer */
    gx_band_params_t band;      /* see gxband.h */
    bool params_are_read_only;  /* true if put_params may not modify this struct */
    gdev_banding_type banding_type;     /* used to force banding or bitmap */
} gdev_space_params;

typedef struct gdev_nupcontrol_s {
        rc_header rc;
        char *nupcontrol_str;           /* NUL termintated string */
} gdev_nupcontrol;

typedef struct gdev_pagelist_s {
        rc_header rc;
        char *Pages;
} gdev_pagelist;

#define dev_t_proc_initialize_device_procs(proc, dev_t)\
  void proc(dev_t *dev)
#define dev_proc_initialize_device_procs(proc)\
  dev_t_proc_initialize_device_procs((proc), gx_device)

struct gx_device_procs_s gx_device_proc_struct(gx_device);
typedef struct gx_device_procs_s gx_device_procs;


/* ---------------- Device procedures ---------------- */

/*
 * Definition of device procedures.
 * Note that the gx_device * argument is not declared const,
 * because many drivers maintain dynamic state in the device structure.
 * Note also that the structure is defined as a template, so that
 * we can instantiate it with device subclasses.
 * Because C doesn't have real templates, we must do this with macros.
 */

/* Define macros for declaring device procedures. */

typedef struct _gs_dev_t gs_dev_t;

#define dev_proc_initialize_device(proc)\
  int proc(gs_dev_t *dev)

#define dev_proc_open_device(proc)\
  int proc(gs_dev_t *dev)

#define dev_t_proc_get_initial_matrix(proc, __dev_t)\
  void proc(gs_dev_t *dev, gs_matrix *pmat)
#define dev_proc_get_initial_matrix(proc)\
  dev_t_proc_get_initial_matrix(proc, gx_device)

#define dev_t_proc_sync_output(proc, __dev_t)\
  int proc(gs_dev_t *dev)
#define dev_proc_sync_output(proc)\
  dev_t_proc_sync_output(proc, gx_device)

#define dev_t_proc_output_page(proc, __dev_t)\
  int proc(gs_dev_t *dev, int num_copies, int flush)
#define dev_proc_output_page(proc)\
  dev_t_proc_output_page(proc, gx_device)

#define dev_t_proc_close_device(proc, dev_t)\
  int proc(gs_dev_t *dev)
#define dev_proc_close_device(proc)\
  dev_t_proc_close_device(proc, gx_device)

#define dev_t_proc_map_rgb_color(proc, dev_t)\
  gx_color_index proc(gs_dev_t *dev, const gx_color_value cv[])
#define dev_proc_map_rgb_color(proc)\
  dev_t_proc_map_rgb_color(proc, gx_device)

#define dev_t_proc_map_color_rgb(proc, dev_t)\
  int proc(gs_dev_t *dev,\
    gx_color_index color, gx_color_value rgb[3])
#define dev_proc_map_color_rgb(proc)\
  dev_t_proc_map_color_rgb(proc, gx_device)

#define dev_proc_fill_rectangle(proc)\
  int proc(gs_dev_t *dev,\
    int x, int y, int width, int height, gx_color_index color)

#define dev_proc_tile_rectangle(proc)\
  int proc(gs_dev_t *dev,\
    const gx_tile_bitmap *tile, int x, int y, int width, int height,\
    gx_color_index color0, gx_color_index color1,\
    int phase_x, int phase_y)

#define dev_t_proc_copy_mono(proc, dev_t)\
  int proc(gs_dev_t *dev,\
    const byte *data, int data_x, int raster, gx_bitmap_id id,\
    int x, int y, int width, int height,\
    gx_color_index color0, gx_color_index color1)
#define dev_proc_copy_mono(proc)\
  dev_t_proc_copy_mono(proc, gx_device)

#define dev_t_proc_copy_color(proc, dev_t)\
  int proc(gs_dev_t *dev,\
    const byte *data, int data_x, int raster, gx_bitmap_id id,\
    int x, int y, int width, int height)
#define dev_proc_copy_color(proc)\
  dev_t_proc_copy_color(proc, gx_device)

                /* OBSOLETED in release 3.66 */

#define dev_t_proc_draw_line(proc, dev_t)\
  int proc(gs_dev_t *dev,\
    int x0, int y0, int x1, int y1, gx_color_index color)
#define dev_proc_draw_line(proc)\
  dev_t_proc_draw_line(proc, gx_device)

                /* Added in release 2.4 */

#define dev_t_proc_get_bits(proc, dev_t)\
  int proc(gs_dev_t *dev,\
    int y, byte *data, byte **actual_data)
#define dev_proc_get_bits(proc)\
  dev_t_proc_get_bits(proc, gx_device)

                /* Added in release 2.4, changed in 2.8, */
                /* renamed in 2.9.6 */

#define dev_t_proc_get_params(proc, dev_t)\
  int proc(gs_dev_t *dev, gs_param_list *plist)
#define dev_proc_get_params(proc)\
  dev_t_proc_get_params(proc, gx_device)

#define dev_t_proc_put_params(proc, dev_t)\
  int proc(gs_dev_t *dev, gs_param_list *plist)
#define dev_proc_put_params(proc)\
  dev_t_proc_put_params(proc, gx_device)

                /* Added in release 2.6 */

#define dev_t_proc_map_cmyk_color(proc, dev_t)\
  gx_color_index proc(gs_dev_t *dev, const gx_color_value cv[])
#define dev_proc_map_cmyk_color(proc)\
  dev_t_proc_map_cmyk_color(proc, gx_device)

#define dev_t_proc_get_xfont_procs(proc, dev_t)\
  const gx_xfont_procs *proc(gs_dev_t *dev)
#define dev_proc_get_xfont_procs(proc)\
  dev_t_proc_get_xfont_procs(proc, gx_device)

                /* Added in release 2.6.1 */

#define dev_t_proc_get_xfont_device(proc, dev_t)\
  gx_device *proc(dev_t *dev)
#define dev_proc_get_xfont_device(proc)\
  dev_t_proc_get_xfont_device(proc, gx_device)

                /* Added in release 2.7.1 */

#define dev_t_proc_map_rgb_alpha_color(proc, dev_t)\
  gx_color_index proc(gs_dev_t *dev,\
    gx_color_value red, gx_color_value green, gx_color_value blue,\
    gx_color_value alpha)
#define dev_proc_map_rgb_alpha_color(proc)\
  dev_t_proc_map_rgb_alpha_color(proc, gx_device)

                /* Added in release 2.8.1 */

#define dev_t_proc_get_page_device(proc, dev_t)\
  gx_device *proc(gs_dev_t *dev)
#define dev_proc_get_page_device(proc)\
  dev_t_proc_get_page_device(proc, gx_device)

                /* Added in release 3.20, OBSOLETED in 5.65 */

#define dev_t_proc_get_alpha_bits(proc, dev_t)\
  int proc(gs_dev_t *dev, graphics_object_type type)
#define dev_proc_get_alpha_bits(proc)\
  dev_t_proc_get_alpha_bits(proc, gx_device)

                /* Added in release 3.20 */

#define dev_t_proc_copy_alpha(proc, dev_t)\
  int proc(gs_dev_t *dev, const byte *data, int data_x,\
    int raster, gx_bitmap_id id, int x, int y, int width, int height,\
    gx_color_index color, int depth)
#define dev_proc_copy_alpha(proc)\
  dev_t_proc_copy_alpha(proc, gx_device)

                /* Added in release 3.38 */

#define dev_t_proc_get_band(proc, dev_t)\
  int proc(gs_dev_t *dev, int y, int *band_start)
#define dev_proc_get_band(proc)\
  dev_t_proc_get_band(proc, gx_device)

                /* Added in release 3.44 */

#define dev_t_proc_copy_rop(proc, dev_t)\
  int proc(gs_dev_t *dev,\
    const byte *sdata, int sourcex, uint sraster, gx_bitmap_id id,\
    const gx_color_index *scolors,\
    const gx_tile_bitmap *texture, const gx_color_index *tcolors,\
    int x, int y, int width, int height,\
    int phase_x, int phase_y, gs_logical_operation_t lop)
#define dev_proc_copy_rop(proc)\
  dev_t_proc_copy_rop(proc, gx_device)

                /* Added in release 3.60, changed in 3.68. */

#define dev_t_proc_fill_path(proc, dev_t)\
  int proc(gs_dev_t *dev,\
    const gs_gstate *pgs, gx_path *ppath,\
    const gx_fill_params *params,\
    const gx_drawing_color *pdcolor, const gx_clip_path *pcpath)
#define dev_proc_fill_path(proc)\
  dev_t_proc_fill_path(proc, gx_device)

#define dev_proc_stroke_path(proc)\
  int proc(gs_dev_t *dev,\
    const gs_gstate *pgs, gx_path *ppath,\
    const gx_stroke_params *params,\
    const gx_drawing_color *pdcolor, const gx_clip_path *pcpath)

                /* Added in release 9.22 */

#define dev_proc_fill_stroke_path(proc)\
  int proc(gs_dev_t *dev,\
    const gs_gstate *pgs, gx_path *ppath,\
    const gx_fill_params *fill_params,\
    const gx_drawing_color *pdcolor_fill,\
    const gx_stroke_params *stroke_params,\
    const gx_drawing_color *pdcolor_stroke,\
    const gx_clip_path *pcpath)

                /* Added in release 9.57 */

#define dev_t_proc_lock_pattern(proc, dev_t)\
  int proc(dev_t *dev,\
           gs_gstate *pgs,\
           gs_id pattern_id,\
           int lock)
#define dev_proc_lock_pattern(proc)\
  dev_t_proc_lock_pattern(proc, gx_device)

                /* Added in release 3.60 */

#define dev_t_proc_fill_mask(proc, dev_t)\
  int proc(gs_dev_t *dev,\
    const byte *data, int data_x, int raster, gx_bitmap_id id,\
    int x, int y, int width, int height,\
    const gx_drawing_color *pdcolor, int depth,\
    gs_logical_operation_t lop, const gx_clip_path *pcpath)
#define dev_proc_fill_mask(proc)\
  dev_t_proc_fill_mask(proc, gx_device)

                /* Added in release 3.66, changed in 3.69 */

#define dev_t_proc_fill_trapezoid(proc, __dev_t)\
  int proc(gs_dev_t *dev,\
    const gs_fixed_edge *left, const gs_fixed_edge *right,\
    fixed ybot, fixed ytop, bool swap_axes,\
    const gx_drawing_color *pdcolor, gs_logical_operation_t lop)
#define dev_proc_fill_trapezoid(proc)\
  dev_t_proc_fill_trapezoid(proc, gx_device)

#define dev_t_proc_fill_parallelogram(proc, dev_t)\
  int proc(gs_dev_t *dev,\
    fixed px, fixed py, fixed ax, fixed ay, fixed bx, fixed by,\
    const gx_drawing_color *pdcolor, gs_logical_operation_t lop)
#define dev_proc_fill_parallelogram(proc)\
  dev_t_proc_fill_parallelogram(proc, gx_device)

#define dev_t_proc_fill_triangle(proc, dev_t)\
  int proc(gs_dev_t *dev,\
    fixed px, fixed py, fixed ax, fixed ay, fixed bx, fixed by,\
    const gx_drawing_color *pdcolor, gs_logical_operation_t lop)
#define dev_proc_fill_triangle(proc)\
  dev_t_proc_fill_triangle(proc, gx_device)

                /* adjustx and adjusty were added in 8.71 to get around a
                 * problem with PCL (Bug 691030). In the fullness of time
                 * hopefully PCL can be fixed to not need them and they can
                 * be removed again. */

#define dev_t_proc_draw_thin_line(proc, dev_t)\
  int proc(gs_dev_t *dev,\
    fixed fx0, fixed fy0, fixed fx1, fixed fy1,\
    const gx_drawing_color *pdcolor, gs_logical_operation_t lop,\
    fixed adjustx, fixed adjusty)
#define dev_proc_draw_thin_line(proc)\
  dev_t_proc_draw_thin_line(proc, gx_device)

                /* Added in release 3.66 (as stubs); */
                /* changed in 3.68; */
                /* begin_image and image_data changed in 4.30, */
                /* begin_image changed in 5.23. */

#define dev_proc_begin_image(proc)\
  int proc(gs_dev_t *dev,\
    const gs_gstate *pgs, const gs_image_t *pim,\
    gs_image_format_t format, const gs_int_rect *prect,\
    const gx_drawing_color *pdcolor, const gx_clip_path *pcpath,\
    gs_memory_t *memory, gx_image_enum_common_t **pinfo)

                /* OBSOLETED in release 5.23 */

#define dev_t_proc_image_data(proc, dev_t)\
  int proc(gs_dev_t *dev,\
    gx_image_enum_common_t *info, const byte **planes, int data_x,\
    uint raster, int height)
#define dev_proc_image_data(proc)\
  dev_t_proc_image_data(proc, gx_device)

                /* OBSOLETED in release 5.23 */

#define dev_proc_end_image(proc)\
  int proc(gs_dev_t *dev,\
    gx_image_enum_common_t *info, bool draw_last)

                /* Added in release 3.68 */

#define dev_t_proc_strip_tile_rectangle(proc, dev_t)\
  int proc(gs_dev_t *dev,\
    const gx_strip_bitmap *tiles, int x, int y, int width, int height,\
    gx_color_index color0, gx_color_index color1,\
    int phase_x, int phase_y)
#define dev_proc_strip_tile_rectangle(proc)\
  dev_t_proc_strip_tile_rectangle(proc, gx_device)

#define dev_t_proc_strip_copy_rop(proc, dev_t)\
  int proc(gs_dev_t *dev,\
    const byte *sdata, int sourcex, uint sraster, gx_bitmap_id id,\
    const gx_color_index *scolors,\
    const gx_strip_bitmap *textures, const gx_color_index *tcolors,\
    int x, int y, int width, int height,\
    int phase_x, int phase_y, gs_logical_operation_t lop)
#define dev_proc_strip_copy_rop(proc)\
  dev_t_proc_strip_copy_rop(proc, gx_device)

                /* Added in release 4.20 */

#define dev_t_proc_get_clipping_box(proc, dev_t)\
  void proc(gs_dev_t *dev, gs_fixed_rect *pbox)
#define dev_proc_get_clipping_box(proc)\
  dev_t_proc_get_clipping_box(proc, gx_device)

                /* Added in release 5.20, changed in 5.23 */

#define dev_t_proc_begin_typed_image(proc, dev_t)\
  int proc(gs_dev_t *dev,\
    const gs_gstate *pgs, const gs_matrix *pmat,\
    const gs_image_common_t *pim, const gs_int_rect *prect,\
    const gx_drawing_color *pdcolor, const gx_clip_path *pcpath,\
    gs_memory_t *memory, gx_image_enum_common_t **pinfo)
#define dev_proc_begin_typed_image(proc)\
  dev_t_proc_begin_typed_image(proc, gx_device)

                /* Added in release 5.20 */

#define dev_t_proc_get_bits_rectangle(proc, dev_t)\
  int proc(gs_dev_t *dev, const gs_int_rect *prect,\
    gs_get_bits_params_t *params, gs_int_rect **unread)
#define dev_proc_get_bits_rectangle(proc)\
  dev_t_proc_get_bits_rectangle(proc, gx_device)

#define dev_t_proc_map_color_rgb_alpha(proc, dev_t)\
  int proc(gs_dev_t *dev,\
    gx_color_index color, gx_color_value rgba[4])
#define dev_proc_map_color_rgb_alpha(proc)\
  dev_t_proc_map_color_rgb_alpha(proc, gx_device)

#define dev_proc_composite(proc)\
  int proc(gs_dev_t *dev,\
    gx_device **pcdev, const gs_composite_t *pcte,\
    gs_gstate *pgs, gs_memory_t *memory, gx_device *cdev)

                /* Added in release 5.23 */

#define dev_t_proc_get_hardware_params(proc, dev_t)\
  int proc(gs_dev_t *dev, gs_param_list *plist)
#define dev_proc_get_hardware_params(proc)\
  dev_t_proc_get_hardware_params(proc, gx_device)

                /* Added in release 5.24 */

     /* ... text_begin ... see gstext.h for definition */

                /* Added in release 6.23 */

#define dev_t_proc_finish_copydevice(proc, dev_t)\
  int proc(gs_dev_t *dev, const gx_device *from_dev)
#define dev_proc_finish_copydevice(proc)\
  dev_t_proc_finish_copydevice(proc, gx_device)

                /* Added in release 6.61 (raph) */

/*
  This area of the transparency facilities is in flux.  Here is a proposal
  for extending the driver interface.
*/

/*
  Push the current transparency state (*ppts) onto the associated stack,
  and set *ppts to a new transparency state of the given dimension.  The
  transparency state may copy some or all of the gs_gstate, such as the
  current alpha and/or transparency mask values, and definitely copies the
  parameters.
*/
#define dev_t_proc_begin_transparency_group(proc, dev_t)\
  int proc(gx_device *dev,\
    const gs_transparency_group_params_t *ptgp,\
    const gs_rect *pbbox,\
    gs_gstate *pgs,\
    gs_memory_t *mem)
#define dev_proc_begin_transparency_group(proc)\
  dev_t_proc_begin_transparency_group(proc, gx_device)

/*
  End a transparency group: blend the top element of the transparency
  stack, which must be a group, into the next-to-top element, popping the
  stack.  If the stack only had a single element, blend into the device
  output.  Set *ppts to 0 iff the stack is now empty.  If end_group fails,
  the stack is *not* popped.
*/
#define dev_t_proc_end_transparency_group(proc, dev_t)\
  int proc(gx_device *dev,\
    gs_gstate *pgs)
#define dev_proc_end_transparency_group(proc)\
  dev_t_proc_end_transparency_group(proc, gx_device)

/*
  Push the transparency state and prepare to render a transparency mask.
  This is similar to begin_transparency_group except that it only
  accumulates coverage values, not full pixel values.
*/
#define dev_t_proc_begin_transparency_mask(proc, dev_t)\
  int proc(gx_device *dev,\
    const gx_transparency_mask_params_t *ptmp,\
    const gs_rect *pbbox,\
    gs_gstate *pgs,\
    gs_memory_t *mem)
#define dev_proc_begin_transparency_mask(proc)\
  dev_t_proc_begin_transparency_mask(proc, gx_device)

/*
  Store a pointer to the rendered transparency mask into *pptm, popping the
  stack like end_group.  Normally, the client will follow this by using
  rc_assign to store the rendered mask into pgs->{opacity,shape}.mask.  If
  end_mask fails, the stack is *not* popped.
*/
#define dev_t_proc_end_transparency_mask(proc, dev_t)\
  int proc(gx_device *dev,\
    gs_gstate *pgs)
#define dev_proc_end_transparency_mask(proc)\
  dev_t_proc_end_transparency_mask(proc, gx_device)

/*
  This will clean up the entire device allocations as something went
  wrong in the middle of reading in the source content while we are dealing with
  a transparency device.
*/
#define dev_t_proc_discard_transparency_layer(proc, dev_t)\
  int proc(gx_device *dev,\
    gs_gstate *pgs)
#define dev_proc_discard_transparency_layer(proc)\
  dev_t_proc_discard_transparency_layer(proc, gx_device)

     /* (end of transparency driver interface extensions) */

/* (start of DeviceN color support) */
/*
 * The following macros are defined in gxcmap.h
 *
 * dev_t_proc_get_color_mapping_procs
 * dev_proc_get_color_mapping_procs
 * dev_t_proc_get_color_comp_index
 * dev_proc_get_color_comp_index
 * dev_t_proc_encode_color
 * dev_proc_encode_color
 * dev_t_proc_decode_color
 * dev_proc_decode_color
 */

/* (end of DeviceN color support) */

/*
  Pattern management for high level devices.
  Now we need it for PatternType 1 only.
  Return codes :
  1 - the device handles high level patterns.
  0 - the device needs low level pattern tiles.
  <0 - error.

  THIS IS NOW DEPRECATED, AND UNUSED. DO NOT USE THIS. USE THIS NOT. THIS
  IS NOT TO BE USED.

  Pattern Management is now done using the dev_spec_op calls.
*/

/* High level device support. */
typedef enum {
    pattern_manage__can_accum,
    pattern_manage__start_accum,
    pattern_manage__finish_accum,
    pattern_manage__load,
    pattern_manage__shading_area,
    pattern_manage__is_cpath_accum,
    pattern_manage__shfill_doesnt_need_path,
    pattern_manage__handles_clip_path
} pattern_manage_t;

#define dev_proc_pattern_manage(proc)\
  int proc(gs_dev_t*, gx_bitmap_id id,\
                gs_pattern1_instance_t *pinst, pattern_manage_t function)

/*
  Fill rectangle with a high level color.
  Return rangecheck, if the device can't handle the high level color.

  The graphics library calls this function with degenerate (widths=0)
  rectangles, to know whether the device can handle a rectangle with
  the high level color. The device should skip such rectangles returning
  a proper code.

  Currently this function is used with gs_rectfill and gs_fillpage.  It is
  also used for the handling of the devn color type for supporting
  large number of spot colorants to planar separation devices.

*/

#define dev_t_proc_fill_rectangle_hl_color(proc, dev_t)\
  int proc(dev_t *dev, const gs_fixed_rect *rect, \
        const gs_gstate *pgs, const gx_drawing_color *pdcolor, \
        const gx_clip_path *pcpath)
#define dev_proc_fill_rectangle_hl_color(proc)\
  dev_t_proc_fill_rectangle_hl_color(proc, gx_device)

/*
  Include a color space into the output.
  This function is used to include DefaultGray, DefaultRGB,
  DefaultCMYK into PDF, PS, EPS output.
  Low level devices should ignore this call.
*/

#define dev_t_proc_include_color_space(proc, dev_t)\
  int proc(dev_t *dev, gs_color_space *cspace, const byte *res_name, int name_length)
#define dev_proc_include_color_space(proc)\
  dev_t_proc_include_color_space(proc, gx_device)

                /* Shading support. */

typedef struct gs_fill_attributes_s {
      const gs_fixed_rect *clip;
      bool swap_axes;
      const gx_device_halftone *ht; /* Reserved for possible use in future. */
      gs_logical_operation_t lop; /* Reserved for possible use in future. */
      fixed ystart, yend; /* Only for X-independent gradients. Base coordinates of the gradient. */
      patch_fill_state_t *pfs; /* For gx_fill_triangle_small. Clients must not change. */
} gs_fill_attributes;

/* Fill a linear color scanline. */

#define dev_t_proc_fill_linear_color_scanline(proc, dev_t)\
  int proc(dev_t *dev, const gs_fill_attributes *fa,\
        int i, int j, int w, /* scanline coordinates and width */\
        const frac31 *c0, /* initial color for the pixel (i,j), the integer part */\
        const int32_t *c0_f, /* initial color for the pixel (i,j), the fraction part numerator */\
        const int32_t *cg_num, /* color gradient numerator */\
        int32_t cg_den /* color gradient denominator */)
#define dev_proc_fill_linear_color_scanline(proc)\
  dev_t_proc_fill_linear_color_scanline(proc, gx_device)

/* Fill a linear color trapezoid. */
/* The server assumes a strongly linear color,
   i.e. it can ignore any of c0, c1, c2, c3. */
/* [p0 : p1] - left edge, from bottom to top.
   [p2 : p3] - right edge, from bottom to top.
   The filled area is within Y-spans of both edges. */
/* Either (c0 and c1) or (c2 and c3) may be NULL.
   In this case the color doesn't depend on X (on Y if fa->swap_axes).
   In this case the base coordinates for the color gradient
   may be unequal to p0, p1, p2, p3, and must be provided/taken
   in/from fa->ystart, fa->yend.
   The return value 0 is not allowed in this case. */
/* Return values :
  1 - success;
  0 - Too big. The area isn't filled. The client must decompose the area.
  <0 - error.
 */

#define dev_t_proc_fill_linear_color_trapezoid(proc, dev_t)\
  int proc(dev_t *dev, const gs_fill_attributes *fa,\
        const gs_fixed_point *p0, const gs_fixed_point *p1,\
        const gs_fixed_point *p2, const gs_fixed_point *p3,\
        const frac31 *c0, const frac31 *c1,\
        const frac31 *c2, const frac31 *c3)
#define dev_proc_fill_linear_color_trapezoid(proc)\
  dev_t_proc_fill_linear_color_trapezoid(proc, gx_device)

/* Fill a linear color triangle. */
/* Return values :
  1 - success;
  0 - Too big. The area isn't filled. The client must decompose the area.
  <0 - error.
 */

#define dev_t_proc_fill_linear_color_triangle(proc, dev_t)\
  int proc(dev_t *dev, const gs_fill_attributes *fa,\
        const gs_fixed_point *p0, const gs_fixed_point *p1,\
        const gs_fixed_point *p2,\
        const frac31 *c0, const frac31 *c1, const frac31 *c2)
#define dev_proc_fill_linear_color_triangle(proc)\
  dev_t_proc_fill_linear_color_triangle(proc, gx_device)

/*
 * Update the equivalent colors for spot colors in a color space.  The default
 * procedure does nothing.  However this routine provides a method for devices
 * to determine an equivalent color for a spot color.  See comments at the
 * start of src/gsequivc.c.
 */
#if GHOSTSCRIPT_MAJOR_VERSION < 10
#define dev_t_proc_update_spot_equivalent_colors(proc, dev_t)\
  int proc(dev_t *dev, const gs_gstate * pgs)
#define dev_proc_update_spot_equivalent_colors(proc)\
  dev_t_proc_update_spot_equivalent_colors(proc, gx_device)
#else
#define dev_t_proc_update_spot_equivalent_colors(proc, dev_t)\
  int proc(dev_t *dev, const gs_gstate * pgs, const gs_color_space *pcs)
#define dev_proc_update_spot_equivalent_colors(proc)\
  dev_t_proc_update_spot_equivalent_colors(proc, gx_device)
#endif

/*
 * return a pointer to the devn_params section of a device.  Return NULL
 * if this field is not present within the device.
 */
typedef struct gs_devn_params_s gs_devn_params;

#define dev_proc_ret_devn_params(proc)\
  gs_devn_params * proc(const dev_t *dev)

/*
 * Erase page.
 */

#define dev_proc_fillpage(proc)\
  int proc(gx_device *dev, gs_gstate * pgs, gx_device_color *pdevc)

#define dev_proc_push_transparency_state(proc)\
  int proc(gx_device *dev,\
    gs_gstate *pgs)

#define dev_t_proc_pop_transparency_state(proc, dev_t)\
  int proc(gx_device *dev,\
    gs_gstate *pgs)
#define dev_proc_pop_transparency_state(proc)\
  dev_t_proc_pop_transparency_state(proc, gx_device)

#define dev_t_proc_put_image(proc, dev_t)\
  int proc(gx_device *dev, gx_device *mdev,  const byte **buffers, int num_chan, int x, int y,\
            int width, int height, int row_stride,\
            int alpha_plane_index, int tag_plane_index)
#define dev_proc_put_image(proc)\
  dev_t_proc_put_image(proc, gx_device)

#define dev_proc_dev_spec_op(proc)\
  int proc(gs_dev_t *dev, int op, void *data, int datasize)

#define dev_t_proc_copy_planes(proc, dev_t)\
  int proc(dev_t *dev,\
    const byte *data, int data_x, int raster, gx_bitmap_id id,\
    int x, int y, int width, int height, int plane_height)
#define dev_proc_copy_planes(proc)\
  dev_t_proc_copy_planes(proc, gx_device)

#define dev_proc_get_profile(proc)\
  int proc(const gs_dev_t *dev, cmm_dev_profile_t **dev_profile)

#define dev_t_proc_set_graphics_type_tag(proc, dev_t)\
  void proc(dev_t *dev, gs_graphics_type_tag_t)
#define dev_proc_set_graphics_type_tag(proc)\
  dev_t_proc_set_graphics_type_tag(proc, gx_device)

#define dev_t_proc_strip_copy_rop2(proc, dev_t)\
  int proc(dev_t *dev,\
    const byte *sdata, int sourcex, uint sraster, gx_bitmap_id id,\
    const gx_color_index *scolors,\
    const gx_strip_bitmap *textures, const gx_color_index *tcolors,\
    int x, int y, int width, int height,\
    int phase_x, int phase_y, gs_logical_operation_t lop,\
    uint planar_height)
#define dev_proc_strip_copy_rop2(proc)\
  dev_t_proc_strip_copy_rop2(proc, gx_device)

#define dev_t_proc_strip_tile_rect_devn(proc, dev_t)\
  int proc(dev_t *dev,\
    const gx_strip_bitmap *tiles, int x, int y, int width, int height,\
    const gx_drawing_color *pdcolor0, const gx_drawing_color *pdcolor1,\
    int phase_x, int phase_y)
#define dev_proc_strip_tile_rect_devn(proc)\
  dev_t_proc_strip_tile_rect_devn(proc, gx_device)

#define dev_t_proc_copy_alpha_hl_color(proc, dev_t)\
  int proc(dev_t *dev, const byte *data, int data_x,\
    int raster, gx_bitmap_id id, int x, int y, int width, int height,\
    const gx_drawing_color *pdcolor, int depth)
#define dev_proc_copy_alpha_hl_color(proc)\
  dev_t_proc_copy_alpha_hl_color(proc, gx_device)

typedef struct gx_process_page_options_s gx_process_page_options_t;

struct gx_process_page_options_s
{
    int (*init_buffer_fn)(void *arg, gx_device *dev, gs_memory_t *memory, int w, int h, void **buffer);
    void (*free_buffer_fn)(void *arg, gx_device *dev, gs_memory_t *memory, void *buffer);
    int (*process_fn)(void *arg, gx_device *dev, gx_device *bdev, const gs_int_rect *rect, void *buffer);
    int (*output_fn)(void *arg, gx_device *dev, void *buffer);
    void *arg;
    int options; /* A mask of GX_PROCPAGE_... options bits */
};

/* If GX_PROCPAGE_BOTTOM_UP, then we run from band n-1 to band 0, rather than
 * 0 to n-1. */
#define GX_PROCPAGE_BOTTOM_UP 1

#define dev_t_proc_process_page(proc, dev_t)\
  int proc(dev_t *dev, gx_process_page_options_t *options)
#define dev_proc_process_page(proc)\
  dev_t_proc_process_page(proc, gx_device)

typedef enum  {
    transform_pixel_region_begin = 0,
    transform_pixel_region_data_needed = 1,
    transform_pixel_region_process_data = 2,
    transform_pixel_region_end = 3
} transform_pixel_region_reason;

typedef struct {
    void *state;
    union {
        struct {
            const gs_int_rect *clip;
            int w; /* source width */
            int h; /* source height */
            int spp;
            const gx_dda_fixed_point *pixels; /* DDA to enumerate the destination positions of pixels across a row */
            const gx_dda_fixed_point *rows; /* DDA to enumerate the starting position of each row */
            gs_logical_operation_t lop;
        } init;
        struct {
            const unsigned char *buffer[GX_DEVICE_COLOR_MAX_COMPONENTS];
            int data_x;
            gx_cmapper_t *cmapper;
            const gs_gstate *pgs;
        } process_data;
    } u;
} transform_pixel_region_data;

#define dev_proc_transform_pixel_region(proc)\
  int proc(gs_dev_t*, transform_pixel_region_reason, transform_pixel_region_data*)

typedef struct _gs_dev_t gs_dev_t;

/* Copied from the macro gx_device_proc_struct in base/gxdevcli.h */
typedef struct{
#if GS_CHECK_VERSION(9,56,0)
    dev_proc_initialize_device((*initialize_device));
#endif
    dev_proc_open_device((*open_device));
    dev_proc_get_initial_matrix((*get_initial_matrix));
    dev_proc_sync_output((*sync_output));
    dev_proc_output_page((*output_page));
    dev_proc_close_device((*close_device));
    dev_proc_map_rgb_color((*map_rgb_color));
    dev_proc_map_color_rgb((*map_color_rgb));
    dev_proc_fill_rectangle((*fill_rectangle));
    dev_proc_copy_mono((*copy_mono));
    dev_proc_copy_color((*copy_color));
#if GS_CHECK_VERSION(2,4,0)
    dev_proc_get_params((*get_params));
    dev_proc_put_params((*put_params));
#endif
#if GS_CHECK_VERSION(2,6,0)
    dev_proc_map_cmyk_color((*map_cmyk_color));
#endif
    dev_proc_get_page_device((*get_page_device));
    dev_proc_get_alpha_bits((*get_alpha_bits));
    dev_proc_copy_alpha((*copy_alpha));
    dev_proc_fill_path((*fill_path));
    dev_proc_stroke_path((*stroke_path));
    dev_proc_fill_mask((*fill_mask));
    dev_proc_fill_trapezoid((*fill_trapezoid));
    dev_proc_fill_parallelogram((*fill_parallelogram));
    dev_proc_fill_triangle((*fill_triangle));
    dev_proc_draw_thin_line((*draw_thin_line));
    dev_proc_strip_tile_rectangle((*strip_tile_rectangle));
#if !GS_CHECK_VERSION(9,56,0)
    dev_proc_begin_image((*begin_image));
    dev_proc_image_data((*image_data));
    dev_proc_end_image((*end_image));
#endif
    dev_proc_get_clipping_box((*get_clipping_box));
    dev_proc_begin_typed_image((*begin_typed_image));
    dev_proc_get_bits_rectangle((*get_bits_rectangle));
#if GS_CHECK_VERSION(9,56,0)
    dev_proc_composite((*composite));
#endif
    dev_proc_get_hardware_params((*get_hardware_params));
    dev_proc_text_begin((*text_begin));
#if GS_CHECK_VERSION(6,61,0)
    dev_proc_begin_transparency_group((*begin_transparency_group));
    dev_proc_end_transparency_group((*end_transparency_group));
    dev_proc_begin_transparency_mask((*begin_transparency_mask));
#endif
#if GS_CHECK_VERSION(9,56,0)
    dev_proc_end_transparency_mask((*end_transparency_mask));
#endif
    dev_proc_discard_transparency_layer((*discard_transparency_layer));
    dev_proc_get_color_mapping_procs((*get_color_mapping_procs));
    dev_proc_get_color_comp_index((*get_color_comp_index));
    dev_proc_encode_color((*encode_color));
    dev_proc_decode_color((*decode_color));
    dev_proc_fill_rectangle_hl_color((*fill_rectangle_hl_color));
    dev_proc_include_color_space((*include_color_space));
    dev_proc_fill_linear_color_scanline((*fill_linear_color_scanline));
    dev_proc_fill_linear_color_trapezoid((*fill_linear_color_trapezoid));
    dev_proc_fill_linear_color_triangle((*fill_linear_color_triangle));
    dev_proc_update_spot_equivalent_colors((*update_spot_equivalent_colors));
    dev_proc_ret_devn_params((*ret_devn_params));
    dev_proc_fillpage((*fillpage));
    dev_proc_push_transparency_state((*push_transparency_state));
    dev_proc_pop_transparency_state((*pop_transparency_state));
    dev_proc_put_image((*put_image));
    dev_proc_dev_spec_op((*dev_spec_op));
    dev_proc_copy_planes((*copy_planes));
    dev_proc_get_profile((*get_profile));
    dev_proc_set_graphics_type_tag((*set_graphics_type_tag));
    dev_proc_strip_copy_rop2((*strip_copy_rop2));
    dev_proc_strip_tile_rect_devn((*strip_tile_rect_devn));
    dev_proc_copy_alpha_hl_color((*copy_alpha_hl_color));
    dev_proc_process_page((*process_page));
    dev_proc_transform_pixel_region((*transform_pixel_region));
#if GS_CHECK_VERSION(9,22,0)
    dev_proc_fill_stroke_path((*fill_stroke_path));
#endif
#if GS_CHECK_VERSION(9,57,0)
    dev_proc_lock_pattern((*lock_pattern));
#endif
}dev_procs_t;

/* Copied from the macro gx_device_common in base/gxdevcli.h */
struct _gs_dev_t{
    int params_size;                /* OBSOLETE if stype != 0: */
                                    /* size of this structure */
    dev_proc_initialize_device_procs(*initialize_device_procs);
                                    /* initialize_device_procs */
    const char *dname;              /* the device name */
    gs_memory_t *memory;            /* (0 iff static prototype) */
    gs_memory_type_ptr_t stype;     /* memory manager structure type, */
                                    /* may be 0 if static prototype */
    bool stype_is_dynamic;          /* if true, free the stype when */
                                    /* freeing the device */
    void (*finalize)(gx_device *);  /* finalization to execute */
                                    /* before closing device, if any */
    rc_header rc;                   /* reference count from gstates */
                                    /* and targets, +1 if retained */
    bool retained;                  /* true if retained */
    gx_device *parent;
    gx_device *child;
    void *subclass_data;            /* Must be immovable, non-GC memory, used to store subclass data */
    gdev_pagelist *PageList;
    bool is_open;                   /* true if device has been opened */
    int max_fill_band;              /* limit on band size for fill, */
                                    /* must be 0 or a power of 2 */
                                    /* (see gdevabuf.c for more info) */
    gx_device_color_info color_info;        /* color information */
    gx_device_cached_colors_t cached_colors;
    int width;                      /* width in pixels */
    int height;                     /* height in pixels */
    int pad;                        /* pad to use for buffers; 0 for default */
    int log2_align_mod;             /* align to use for buffers; 0 for default */
    int is_planar;                  /* 1 planar, 0 for chunky */
    int LeadingEdge;                /* see below */
    float MediaSize[2];             /* media dimensions in points */
    float ImagingBBox[4];           /* imageable region in points */
    bool ImagingBBox_set;
    float HWResolution[2];          /* resolution, dots per inch */
    float Margins[2];               /* offset of physical page corner */
                                    /* from device coordinate (0,0), */
                                    /* in units given by HWResolution */
    float HWMargins[4];             /* margins around imageable area, */
                                    /* in default user units ("points") */
    int FirstPage;
    int LastPage;
    bool PageHandlerPushed;         /* Handles FirstPage and LastPage operations */
    bool DisablePageHandler;        /* Can be set by the interpreter if it will process FirstPage and LastPage itself */
    int ObjectFilter;               /* Bit field for which object filters to apply */
    bool ObjectHandlerPushed;       /* Handles filtering of objects to devices */
//#if GS_CHECK_VERSION(10,0)
    gdev_nupcontrol *NupControl;
    bool NupHandlerPushed;          /* Handles Nup operations */
//#endif
    long PageCount;                 /* number of pages written */
    long ShowpageCount;             /* number of calls on showpage */
    int NumCopies;
    bool NumCopies_set;
    bool IgnoreNumCopies;           /* if true, force num_copies = 1 */
    bool UseCIEColor;               /* for PS LL3 */
    bool LockSafetyParams;          /* If true, prevent unsafe changes */
    long band_offset_x;             /* offsets of clist band base to (mem device) buffer */
    long band_offset_y;             /* for rendering that is phase sensitive (old wtsimdi) */
    bool BLS_force_memory;
    gx_stroked_gradient_recognizer_t sgr;
    size_t MaxPatternBitmap;        /* Threshold for switching to pattern_clist mode */
    bool page_uses_transparency;    /* PDF 1.4 transparency is used. */
//#if GHOSTSCRIPT_MAJOR_VERSION >= 10
    bool page_uses_overprint;       /* overprint is used. */
//#endif
    gdev_space_params space_params;
    cmm_dev_profile_t *icc_struct;  /* object dependent profiles */
    gs_graphics_type_tag_t   graphics_type_tag;   /* e.g. vector, image or text */
    int interpolate_control;      /* default 1 (use image /Interpolate value), 0 is NOINTERPOLATE. */
                                  /* > 1 limits interpolation, < 0 forces interpolation */
    gx_page_device_procs page_procs;    /* must be last */
    /* end of std_device_body */
    dev_procs_t procs;                  /* object procedures */
};


/* A generic device procedure record. */
struct gx_device_procs_s gx_device_proc_struct(gx_device);

#if 0
/* copied from base/gxdevcli.h */
/* A generic forwarding device. */
typedef struct gx_device_forward_s {
    gx_device_forward_common;
} gx_device_forward;
#endif

/* Perform a callout to registered handlers from the device. */
int gx_callout(gs_dev_t*, int id, int size, void *data);

/* ---------------------------------------------------------------------- */
/* U.S. letter paper (8.5" x 11"). */
#define DEFAULT_WIDTH_10THS_US_LETTER 85
#define DEFAULT_HEIGHT_10THS_US_LETTER 110
/* A4 paper (210mm x 297mm), we use 595pt x 842pt. */
#define DEFAULT_WIDTH_10THS_A4 82.6389f
#define DEFAULT_HEIGHT_10THS_A4 116.9444f
/* Choose a default.  A4 may be set in the makefile. */
#ifdef A4
# define DEFAULT_WIDTH_10THS DEFAULT_WIDTH_10THS_A4
# define DEFAULT_HEIGHT_10THS DEFAULT_HEIGHT_10THS_A4
#else
# define DEFAULT_WIDTH_10THS DEFAULT_WIDTH_10THS_US_LETTER
# define DEFAULT_HEIGHT_10THS DEFAULT_HEIGHT_10THS_US_LETTER
#endif

/* ---------------- Default implementations ---------------- */

/* Default implementations of optional procedures. */
/* Note that the default map_xxx_color routines assume white_on_black. */
dev_proc_open_device(gx_default_open_device);
dev_proc_get_initial_matrix(gx_default_get_initial_matrix);
dev_proc_get_initial_matrix(gx_upright_get_initial_matrix);
dev_proc_sync_output(gx_default_sync_output);
dev_proc_output_page(gx_default_output_page);
dev_proc_close_device(gx_default_close_device);
dev_proc_map_rgb_color(gx_default_w_b_map_rgb_color);
dev_proc_map_color_rgb(gx_default_w_b_map_color_rgb);
#define gx_default_map_rgb_color gx_default_w_b_map_rgb_color
#define gx_default_map_color_rgb gx_default_w_b_map_color_rgb
dev_proc_tile_rectangle(gx_default_tile_rectangle);
dev_proc_copy_mono(gx_default_copy_mono);
dev_proc_copy_color(gx_default_copy_color);
dev_proc_draw_line(gx_default_draw_line);
dev_proc_get_bits(gx_no_get_bits);      /* gives error */
dev_proc_get_bits(gx_default_get_bits);
dev_proc_get_params(gx_default_get_params);
dev_proc_put_params(gx_default_put_params);
dev_proc_map_cmyk_color(gx_default_map_cmyk_color);
dev_proc_get_xfont_procs(gx_default_get_xfont_procs);
dev_proc_get_xfont_device(gx_default_get_xfont_device);
dev_proc_map_rgb_alpha_color(gx_default_map_rgb_alpha_color);
dev_proc_get_page_device(gx_default_get_page_device);   /* returns NULL */
dev_proc_get_page_device(gx_page_device_get_page_device);       /* returns dev */
dev_proc_get_alpha_bits(gx_default_get_alpha_bits);
dev_proc_copy_alpha(gx_no_copy_alpha);  /* gives error */
dev_proc_copy_alpha(gx_default_copy_alpha);
dev_proc_get_band(gx_default_get_band);
dev_proc_copy_rop(gx_no_copy_rop);      /* gives error */
dev_proc_copy_rop(gx_default_copy_rop);
dev_proc_fill_path(gx_default_fill_path);
dev_proc_stroke_path(gx_default_stroke_path);
dev_proc_fill_mask(gx_default_fill_mask);
dev_proc_fill_trapezoid(gx_default_fill_trapezoid);
dev_proc_fill_parallelogram(gx_default_fill_parallelogram);
dev_proc_fill_triangle(gx_default_fill_triangle);
dev_proc_draw_thin_line(gx_default_draw_thin_line);
dev_proc_begin_image(gx_default_begin_image);
dev_proc_image_data(gx_default_image_data);
dev_proc_end_image(gx_default_end_image);
dev_proc_strip_tile_rectangle(gx_default_strip_tile_rectangle);
dev_proc_strip_copy_rop(gx_no_strip_copy_rop);  /* gives error */
dev_proc_strip_copy_rop(gx_default_strip_copy_rop);
dev_proc_get_clipping_box(gx_default_get_clipping_box);
dev_proc_get_clipping_box(gx_get_largest_clipping_box);
dev_proc_begin_typed_image(gx_default_begin_typed_image);
dev_proc_get_bits_rectangle(gx_no_get_bits_rectangle);  /* gives error */
dev_proc_get_bits_rectangle(gx_default_get_bits_rectangle);
dev_proc_map_color_rgb_alpha(gx_default_map_color_rgb_alpha);
dev_proc_composite(gx_no_create_compositor);
/* default is for ordinary "leaf" devices, null is for */
/* devices that only care about coverage and not contents. */
dev_proc_composite(gx_default_create_compositor);
dev_proc_composite(gx_null_create_compositor);
dev_proc_get_hardware_params(gx_default_get_hardware_params);
dev_proc_text_begin(gx_default_text_begin);
dev_proc_finish_copydevice(gx_default_finish_copydevice);
dev_proc_dev_spec_op(gx_default_dev_spec_op);
dev_proc_fill_rectangle_hl_color(gx_default_fill_rectangle_hl_color);
dev_proc_include_color_space(gx_default_include_color_space);
dev_proc_fill_linear_color_scanline(gx_default_fill_linear_color_scanline);
dev_proc_fill_linear_color_scanline(gx_hl_fill_linear_color_scanline);
dev_proc_fill_linear_color_trapezoid(gx_default_fill_linear_color_trapezoid);
dev_proc_fill_linear_color_triangle(gx_default_fill_linear_color_triangle);
dev_proc_update_spot_equivalent_colors(gx_default_update_spot_equivalent_colors);
dev_proc_ret_devn_params(gx_default_ret_devn_params);
dev_proc_fillpage(gx_default_fillpage);
dev_proc_get_profile(gx_default_get_profile);
dev_proc_set_graphics_type_tag(gx_default_set_graphics_type_tag);
dev_proc_strip_copy_rop2(gx_default_strip_copy_rop2);
dev_proc_strip_tile_rect_devn(gx_default_strip_tile_rect_devn);
dev_proc_copy_alpha_hl_color(gx_default_copy_alpha_hl_color);
dev_proc_process_page(gx_default_process_page);
dev_proc_transform_pixel_region(gx_default_transform_pixel_region);
dev_proc_fill_stroke_path(gx_default_fill_stroke_path);
dev_proc_begin_transparency_group(gx_default_begin_transparency_group);
dev_proc_end_transparency_group(gx_default_end_transparency_group);
dev_proc_begin_transparency_mask(gx_default_begin_transparency_mask);
dev_proc_end_transparency_mask(gx_default_end_transparency_mask);
dev_proc_discard_transparency_layer(gx_default_discard_transparency_layer);
//dev_proc_pattern_manage(gx_default_pattern_manage);
dev_proc_push_transparency_state(gx_default_push_transparency_state);
dev_proc_pop_transparency_state(gx_default_pop_transparency_state);
dev_proc_put_image(gx_default_put_image);
dev_proc_copy_alpha_hl_color(gx_default_no_copy_alpha_hl_color);
dev_proc_copy_planes(gx_default_copy_planes);

/* BACKWARD COMPATIBILITY */
#define gx_non_imaging_create_compositor gx_null_create_compositor

/* Color mapping routines for black-on-white, gray scale, true RGB, */
/* true CMYK, and 1-bit CMYK color. */
dev_proc_map_rgb_color(gx_default_b_w_map_rgb_color);
dev_proc_map_color_rgb(gx_default_b_w_map_color_rgb);
dev_proc_map_rgb_color(gx_default_gray_map_rgb_color);
dev_proc_map_color_rgb(gx_default_gray_map_color_rgb);
dev_proc_map_color_rgb(gx_default_rgb_map_color_rgb);
#define gx_default_cmyk_map_cmyk_color cmyk_8bit_map_cmyk_color /*see below*/
/*
 * The following are defined as "standard" color mapping procedures
 * that can be propagated through device pipelines and that color
 * processing code can test for.
 */
dev_proc_map_rgb_color(gx_default_rgb_map_rgb_color);
dev_proc_map_cmyk_color(cmyk_1bit_map_cmyk_color);
dev_proc_map_color_rgb(cmyk_1bit_map_color_rgb);
#if 0
dev_proc_decode_color(cmyk_1bit_map_color_cmyk);
dev_proc_map_cmyk_color(cmyk_8bit_map_cmyk_color);
dev_proc_map_color_rgb(cmyk_8bit_map_color_rgb);
dev_proc_decode_color(cmyk_8bit_map_color_cmyk);
dev_proc_map_cmyk_color(cmyk_16bit_map_cmyk_color);
dev_proc_decode_color(cmyk_16bit_map_color_cmyk);
dev_proc_encode_color(gx_default_8bit_map_gray_color);
dev_proc_decode_color(gx_default_8bit_map_color_gray);
#endif

/* Default implementations for forwarding devices */
dev_proc_close_device(gx_forward_close_device);
dev_proc_get_initial_matrix(gx_forward_get_initial_matrix);
dev_proc_sync_output(gx_forward_sync_output);
dev_proc_output_page(gx_forward_output_page);
dev_proc_map_rgb_color(gx_forward_map_rgb_color);
dev_proc_map_color_rgb(gx_forward_map_color_rgb);
dev_proc_fill_rectangle(gx_forward_fill_rectangle);
dev_proc_tile_rectangle(gx_forward_tile_rectangle);
dev_proc_copy_mono(gx_forward_copy_mono);
dev_proc_copy_color(gx_forward_copy_color);
dev_proc_get_bits(gx_forward_get_bits);
dev_proc_get_params(gx_forward_get_params);
dev_proc_put_params(gx_forward_put_params);
dev_proc_map_cmyk_color(gx_forward_map_cmyk_color);
dev_proc_get_xfont_procs(gx_forward_get_xfont_procs);
dev_proc_get_xfont_device(gx_forward_get_xfont_device);
dev_proc_map_rgb_alpha_color(gx_forward_map_rgb_alpha_color);
dev_proc_get_page_device(gx_forward_get_page_device);
#define gx_forward_get_alpha_bits gx_default_get_alpha_bits
dev_proc_copy_alpha(gx_forward_copy_alpha);
dev_proc_get_band(gx_forward_get_band);
dev_proc_copy_rop(gx_forward_copy_rop);
dev_proc_fill_path(gx_forward_fill_path);
dev_proc_stroke_path(gx_forward_stroke_path);
dev_proc_fill_mask(gx_forward_fill_mask);
dev_proc_fill_trapezoid(gx_forward_fill_trapezoid);
dev_proc_fill_parallelogram(gx_forward_fill_parallelogram);
dev_proc_fill_triangle(gx_forward_fill_triangle);
dev_proc_draw_thin_line(gx_forward_draw_thin_line);
dev_proc_begin_image(gx_forward_begin_image);
#define gx_forward_image_data gx_default_image_data
#define gx_forward_end_image gx_default_end_image
dev_proc_strip_tile_rectangle(gx_forward_strip_tile_rectangle);
dev_proc_strip_copy_rop(gx_forward_strip_copy_rop);
dev_proc_get_clipping_box(gx_forward_get_clipping_box);
dev_proc_begin_typed_image(gx_forward_begin_typed_image);
dev_proc_get_bits_rectangle(gx_forward_get_bits_rectangle);
dev_proc_map_color_rgb_alpha(gx_forward_map_color_rgb_alpha);
/* There is no forward_create_compositor (see Drivers.htm). */
dev_proc_get_hardware_params(gx_forward_get_hardware_params);
#if 0
dev_proc_text_begin(gx_forward_text_begin);
dev_proc_get_color_mapping_procs(gx_forward_get_color_mapping_procs);
dev_proc_get_color_comp_index(gx_forward_get_color_comp_index);
dev_proc_encode_color(gx_forward_encode_color);
dev_proc_decode_color(gx_forward_decode_color);
dev_proc_dev_spec_op(gx_forward_dev_spec_op);
dev_proc_fill_rectangle_hl_color(gx_forward_fill_rectangle_hl_color);
dev_proc_include_color_space(gx_forward_include_color_space);
dev_proc_fill_linear_color_scanline(gx_forward_fill_linear_color_scanline);
dev_proc_fill_linear_color_trapezoid(gx_forward_fill_linear_color_trapezoid);
dev_proc_fill_linear_color_triangle(gx_forward_fill_linear_color_triangle);
dev_proc_update_spot_equivalent_colors(gx_forward_update_spot_equivalent_colors);
dev_proc_ret_devn_params(gx_forward_ret_devn_params);
dev_proc_fillpage(gx_forward_fillpage);
dev_proc_put_image(gx_forward_put_image);
dev_proc_copy_planes(gx_forward_copy_planes);
dev_proc_create_compositor(gx_forward_create_compositor);
dev_proc_get_profile(gx_forward_get_profile);
dev_proc_set_graphics_type_tag(gx_forward_set_graphics_type_tag);
dev_proc_strip_copy_rop2(gx_forward_strip_copy_rop2);
dev_proc_strip_tile_rect_devn(gx_forward_strip_tile_rect_devn);
dev_proc_copy_alpha_hl_color(gx_forward_copy_alpha_hl_color);
dev_proc_transform_pixel_region(gx_forward_transform_pixel_region);
dev_proc_fill_stroke_path(gx_forward_fill_stroke_path);
#endif

/* ---------------- Implementation utilities ---------------- */
int gx_default_get_param(gx_device *dev, char *Param, void *list);

/* Convert the device procedures to the proper form (see above). */
void gx_device_set_procs(gs_dev_t*);

/* Fill in defaulted procedures in a device procedure record. */
void gx_device_fill_in_procs(gx_device *);
void gx_device_forward_fill_in_procs(gs_dev_t*);

/* Forward the color mapping procedures from a device to its target. */
void gx_device_forward_color_procs(gs_dev_t*);

/*
 * Check if the device's encode_color routine uses 'separable' bit encodings
 * for each colorant.  For more info see the routine's header.
 */
void check_device_separable(gs_dev_t*);
/*
 * Check if the device's encode_color routine uses a pdf14 compatible
 * encoding.  For more info see the routine's header.
 */
void check_device_compatible_encoding(gx_device * dev);
/*
 * If a device has a linear and separable encode color function then
 * set up the comp_bits, comp_mask, and comp_shift fields.
 */
void set_linear_color_bits_mask_shift(gx_device * dev);
/*
 * Copy the color mapping procedures from the target if they are
 * standard ones (saving a level of procedure call at mapping time).
 */
void gx_device_copy_color_procs(gx_device *dev, const gx_device *target);

/* Get the black and white pixel values of a device. */
gx_color_index gx_device_black(gx_device *dev);
#define gx_device_black_inline(dev)\
  ((dev)->cached_colors.black == gx_no_color_index ?\
   gx_device_black(dev) : (dev)->cached_colors.black)
gx_color_index gx_device_white(gx_device *dev);
#define gx_device_white_inline(dev)\
  ((dev)->cached_colors.white == gx_no_color_index ?\
   gx_device_white(dev) : (dev)->cached_colors.white)

/* Clear the black/white pixel cache. */
void gx_device_decache_colors(gx_device *dev);

/*
 * Copy the color-related device parameters back from the target:
 * color_info and color mapping procedures.
 */
void gx_device_copy_color_params(gx_device *dev, const gx_device *target);

/*
 * Copy device parameters back from a target.  This copies all standard
 * parameters related to page size and resolution, plus color_info
 * and (if appropriate) color mapping procedures.
 */
void gx_device_copy_params(gx_device *dev, const gx_device *target);

/*
 * Parse the output file name for a device, recognizing "-" and "|command",
 * and also detecting and validating any %nnd format for inserting the
 * page count.  If a format is present, store a pointer to its last
 * character in *pfmt, otherwise store 0 there.  Note that an empty name
 * is currently allowed.
 */
int gx_parse_output_file_name(gs_parsed_file_name_t *pfn,
                              const char **pfmt, const char *fname,
                              uint len, gs_memory_t *memory);

/*
 * Returns true if the outputfile requests separate pages (contains %d)
 */
bool gx_outputfile_is_separate_pages(const char *fname, gs_memory_t *memory);

/*
 * Open the output file for a device.  Note that if the file name is empty,
 * it may be replaced with the name of a scratch file.
 */
int gx_device_open_output_file(const gx_device * dev, char *fname,
                               bool binary, bool positionable,
                               gp_file ** pfile);

/* Close the output file for a device. */
int gx_device_close_output_file(const gx_device * dev, const char *fname,
                                gp_file *file);

/* Delete the current output file for a device (file must be closed first) */
int gx_device_delete_output_file(const gx_device * dev, const char *fname);

/*
 * Define the number of levels for a colorant above which we do not halftone.
 */
#define MIN_CONTONE_LEVELS 31

/*
 * Determine whether a given device needs to halftone.  Eventually this
 * should take a gs_gstate as an additional argument.
 */
#define gx_device_must_halftone(dev)\
  ((gx_device_has_color(dev) ? (dev)->color_info.max_color :\
    (dev)->color_info.max_gray) < MIN_CONTONE_LEVELS)

/*
 * Do generic work for output_page.  All output_page procedures must call
 * this as the last thing they do, unless an error has occurred earlier.
 */
dev_proc_output_page(gx_finish_output_page);

/*
 * Device procedures that draw into rectangles need to clip the coordinates
 * to the rectangle ((0,0),(dev->width,dev->height)).  The following macros
 * do the clipping.  They assume that the arguments of the procedure are
 * named dev, x, y, w, and h, and may modify the arguments (other than dev).
 *
 * For procedures that fill a region, dev, x, y, w, and h are the only
 * relevant arguments.  For procedures that copy bitmaps, see below.
 *
 * The following group of macros for region-filling procedures clips
 * specific edges of the supplied rectangle, as indicated by the macro name.
 */
#define fit_fill_xy(dev, x, y, w, h)\
  BEGIN\
        if ( (x | y) < 0 ) {\
          if ( x < 0 )\
            w += x, x = 0;\
          if ( y < 0 )\
            h += y, y = 0;\
        }\
  END
#define fit_fill_y(dev, y, h)\
  BEGIN\
        if ( y < 0 )\
          h += y, y = 0;\
  END
#define fit_fill_w(dev, x, w)\
  BEGIN\
        if ( w > (dev)->width - x )\
          w = (dev)->width - x;\
  END
#define fit_fill_h(dev, y, h)\
  BEGIN\
        if ( h > (dev)->height - y )\
          h = (dev)->height - y;\
  END
#define fit_fill_xywh(dev, x, y, w, h)\
  BEGIN\
        fit_fill_xy(dev, x, y, w, h);\
        fit_fill_w(dev, x, w);\
        fit_fill_h(dev, y, h);\
  END
/*
 * Clip all edges, and return from the procedure if the result is empty.
 */
#define fit_fill(dev, x, y, w, h)\
  BEGIN\
        fit_fill_xywh(dev, x, y, w, h);\
        if ( w <= 0 || h <= 0 )\
          return 0;\
  END

/*
 * For driver procedures that copy bitmaps (e.g., copy_mono, copy_color),
 * clipping the destination region also may require adjusting the pointer to
 * the source data.  In addition to dev, x, y, w, and h, the clipping macros
 * for these procedures reference data, data_x, raster, and id; they may
 * modify the values of data, data_x, and id.
 *
 * Clip the edges indicated by the macro name.
 */
#define fit_copy_xyw(dev, data, data_x, raster, id, x, y, w, h)\
  BEGIN\
        if ( (x | y) < 0 ) {\
          if ( x < 0 )\
            w += x, data_x -= x, x = 0;\
          if ( y < 0 )\
            h += y, data -= (int)(y * raster), id = gx_no_bitmap_id, y = 0; \
        }\
        if ( w > (dev)->width - x )\
          w = (dev)->width - x;\
  END
/*
 * Clip all edges, and return from the procedure if the result is empty.
 */
#define fit_copy(dev, data, data_x, raster, id, x, y, w, h)\
  BEGIN\
        fit_copy_xyw(dev, data, data_x, raster, id, x, y, w, h);\
        if ( h > (dev)->height - y )\
          h = (dev)->height - y;\
        if ( w <= 0 || h <= 0 )\
          return 0;\
  END

/* ---------------- Media parameters ---------------- */

/* Define the InputAttributes and OutputAttributes of a device. */
/* The device get_params procedure would call these. */

typedef struct gdev_input_media_s {
    float PageSize[4];          /* nota bene */
    const char *MediaColor;
    float MediaWeight;
    const char *MediaType;
} gdev_input_media_t;

#define gdev_input_media_default_values { 0, 0, 0, 0 }, 0, 0, 0
extern const gdev_input_media_t gdev_input_media_default;

void gdev_input_media_init(gdev_input_media_t * pim);

int gdev_begin_input_media(gs_param_list * mlist, gs_param_dict * pdict,
                           int count);

int gdev_write_input_page_size(int index, gs_param_dict * pdict,
                               double width_points, double height_points);

int gdev_write_input_media(int index, gs_param_dict * pdict,
                           const gdev_input_media_t * pim);

int gdev_end_input_media(gs_param_list * mlist, gs_param_dict * pdict);

typedef struct gdev_output_media_s {
    const char *OutputType;
} gdev_output_media_t;

#define gdev_output_media_default_values 0
extern const gdev_output_media_t gdev_output_media_default;

int gdev_begin_output_media(gs_param_list * mlist, gs_param_dict * pdict,
                            int count);

int gdev_write_output_media(int index, gs_param_dict * pdict,
                            const gdev_output_media_t * pom);

/* Can be called from set user params */
void gx_default_put_icc_dir(gs_param_string *icc_pro, gx_device * dev);

int gdev_end_output_media(gs_param_list * mlist, gs_param_dict * pdict);

void gx_device_request_leadingedge(gx_device *dev, int le_req);

/* ---------------- Device subclassing procedures ---------------- */
int gs_is_pdf14trans_compositor(const gs_composite_t * pct);

#define subclass_common\
    t_dev_proc_create_compositor *saved_compositor_method;\
    gx_device_forward *forwarding_dev

typedef int (t_dev_proc_create_compositor) (gx_device *dev, gx_device **pcdev, const gs_composite_t *pcte, gs_gstate *pgs, gs_memory_t *memory, gx_device *cdev);

typedef struct {
    t_dev_proc_create_compositor *saved_compositor_method;
    gs_dev_t *forwarding_dev;
} generic_subclass_data;


int gx_copy_device_procs(gx_device *dest, gx_device *src, gx_device *prototype);
int gx_device_subclass(gx_device *dev_to_subclass, gx_device *new_prototype, unsigned int private_data_size);
int gx_device_unsubclass(gx_device *dev);
int gx_update_from_subclass(gx_device *dev);
int gx_subclass_create_compositor(gx_device *dev, gx_device **pcdev, const gs_composite_t *pcte,
    gs_gstate *pgs, gs_memory_t *memory, gx_device *cdev);

void gs_lib_register_device(const gs_dev_t*);

#endif /* gs_internal_h */
