/*
 * Copyright (C) 2020  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <debug.h>
#include <stdlib.h>

#include "dcs.h"

/* #define DCS_CHECK_ERROR */
#define DCS_MAX_PARAM_LENGTH 255


struct _dcs_decoder_t{
    const dcs_interface_t *vt;
    void *userdata;

    int incode;
    int command;
    /*
    char param[DCS_MAX_PARAM_LENGTH+1];
    int paramlen;
    */
};


dcs_decoder_t *
dcs_decoder_new(const dcs_interface_t *vt,void *userdata)
{
    dcs_decoder_t *dec = (dcs_decoder_t*)malloc(sizeof(dcs_decoder_t));
    dec->vt = vt;
    dec->userdata = userdata;
    dec->incode = 0;
    return dec;
}


void
dcs_decoder_free(dcs_decoder_t *dec)
{
    free(dec);
}


int
dcs_decoder_write(dcs_decoder_t *dec, int ch)
{
    switch(dec->incode){
    case 0: /* expect ESC */
        if('\e'==ch){
            dec->incode = 1;
        }else{
            /* This character is not part of a DCS command. */
            return 0;
        }
        break;
    case 1: /* got ESC */
        switch(ch){
        case '\\':
            dec->vt->end(dec->userdata, dec->command);
            dec->incode = 0;
            break;
        case 'P':
            dec->incode = 3;
            break;
        default:
            if(dec->vt->write)
                dec->vt->write(dec->userdata, '\e');
            return 0;
        }
        break;
    case 2: /* expect digit */
#ifdef DCS_CHECK_ERROR
        if(ch>='0' && ch<='9'){
            dec->incode = 4;
        }else{
            DEBUG_p("excpected digit, got %c",ch);
        }
        break;
#endif
    case 3: /* expect digit or terminator */
#ifdef DCS_CHECK_ERROR
        if(ch>='0' && ch<='9'){
            dec->incode = 4;
        }else if((ch>='A' && ch<='Z') || (ch>='a' && ch<='z')){
            dec->command = ch;
            dec->vt->begin(dec->userdata, dec->command);
            //dec->paramlen = 0;
            dec->incode = 0;
        }else{
            DEBUG_p("expected digit or terminator, got %c",ch);
        }
        break;
#endif
    case 4: /* expect digit separator or terminator */
        if(ch>='0' && ch<='9'){
            dec->incode = 4;
        }else if(';'==ch){
            dec->incode = 4;
        }else if((ch>='A' && ch<='Z') || (ch>='a' && ch<='z')){
            dec->command = ch;
            dec->vt->begin(dec->userdata, dec->command);
            //dec->paramlen = 0;
            dec->incode = 0;
#ifdef DCS_CHECK_ERROR
        }else{
            DEBUG_p("expected digit, separator or terminator, got %c",ch);
#endif
        }
        break;
    }
    return 1;
}


int
dcs_decoder_writes(dcs_decoder_t *dec, const char *str, size_t len)
{
    const char *end = str+len;
    while(str<end){
        if(!dcs_decoder_write(dec, *str) && dec->vt->write)
            dec->vt->write(dec->userdata, *str);
        ++str;
    }
    return len;
}
