/*
 * Copyright (C) 2020 - 2025  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include "transform.h"


#define transform_get(t) (transform_size(t)>0 ? (cairo_matrix_t*)loa_array_get(&t->matrix_stack, transform_size(t)-1) : NULL)

cairo_matrix_t *
transform_push_identity(transform_t *t)
{
    cairo_matrix_t *m0 = transform_get(t);
    cairo_matrix_t *m = loa_array_append(&t->matrix_stack, 1);
    if(m0){
        *m = *m0;
    }else{
        cairo_matrix_init_identity(m);
    }
    return m;
}


cairo_matrix_t *
transform_push_translate(transform_t *t, double x, double y)
{
    cairo_matrix_t *m0 = transform_get(t);
    cairo_matrix_t *m = loa_array_append(&t->matrix_stack, 1);
    if(m0){
        *m = *m0;
        cairo_matrix_translate(m, x ,y);
    }else{
        cairo_matrix_init_translate(m, x, y);
    }
    return m;
}


cairo_matrix_t *
transform_push_multiply(transform_t *t, cairo_matrix_t *m1)
{
    cairo_matrix_t *m0 = transform_get(t);
    cairo_matrix_t *m = loa_array_append(&t->matrix_stack, 1);
    if(m0){
        cairo_matrix_multiply(m, m1, m0);
    }else{
        *m = *m1;
    }
    return m;
}


void
transform_pop(transform_t *t)
{
    loa_array_delete(&t->matrix_stack, transform_size(t)-1, 1);
}


void
transform_translate(transform_t *t, double x, double y)
{
    cairo_matrix_t *m = transform_get(t);
    if(m)
        cairo_matrix_translate(m, x, y);
}


void
transform_multiply(transform_t *t, cairo_matrix_t *m0)
{
    cairo_matrix_t *m = transform_get(t);
    if(m)
        cairo_matrix_multiply(m, m0, m);
}


void
transform_apply_pos(transform_t *t, double *rx, double *ry)
{
    cairo_matrix_t *m = transform_get(t);
    if(m)
        cairo_matrix_transform_point(m, rx ,ry);
}


void
transform_apply_size(transform_t *t, double *rw, double *rh)
{
    cairo_matrix_t *m = transform_get(t);
    if(m)
        cairo_matrix_transform_distance(m, rw ,rh);
}
