/*
 * Copyright (C) 2017  Teemu Suutari
 * Copyright (C) 2020  Magnus Bergman <magnus@snisurset.net>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * * Redistributions of source code must retain the above copyright notice, this
 *   list of conditions and the following disclaimer.
 *
 * * Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 */
#include "dec_lzbs.h"
#include "streami.h"
#include "streami.h"


LZBSDecompressor::LZBSDecompressor(const Buffer &packedData) :
	XPKDecompressor(packedData)
{
	if (_packedData.size()<1) throw Decompressor::InvalidFormatError();
}

void LZBSDecompressor::decompressImpl(Buffer &rawData,const Buffer &previousData,bool verify)
{
	ForwardInputStream inputStream(_packedData,1,_packedData.size());
	MSBBitReader<ForwardInputStream> bitReader(inputStream);
	auto readBits=[&](uint32_t count)->uint32_t
	{
		return rotateBits(bitReader.readBits8(count),count);
	};

	ForwardOutputStream outputStream(rawData,0,rawData.size());

	uint32_t bits=0,maxBits=uint32_t(_packedData[0]);
	while (!outputStream.eof())
	{
		if (!readBits(1))
		{
			outputStream.writeByte(readBits(8));
		} else {
			uint32_t count=readBits(8)+2;
			if (count==2)
			{
				count=readBits(12);
				if (!count) throw Decompressor::DecompressionError();
				for (uint32_t i=0;i<count;i++)
					outputStream.writeByte(readBits(8));
			} else {
				while (outputStream.getOffset()>=(1U<<bits) && bits<maxBits) bits++;
				uint32_t distance=readBits(bits);

				outputStream.copy(distance,count);
			}
		}
	}
}
