/*
 * Copyright (C) 2020  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include "dec_acca.h"
#include "dec_artm.h"
#include "dec_blzw.h"
#include "dec_cbr0.h"
#include "dec_crm.h"
#include "dec_cyb2.h"
#include "dec_dlta.h"
//#include "dec_dms.h"
#include "dec_fast.h"
#include "dec_fbr2.h"
#include "dec_frle.h"
#include "dec_hfmn.h"
#include "dec_huff.h"
#include "dec_ilzr.h"
#include "dec_imp.h"
#include "dec_lhlb.h"
#include "dec_lin1.h"
#include "dec_lin2.h"
#include "dec_lzbs.h"
#include "dec_lzcb.h"
#include "dec_lzw2.h"
#include "dec_lzw4.h"
#include "dec_lzw5.h"
#include "dec_lzx.h"
#include "dec_mash.h"
//#include "dec_mmcm.h"
#include "dec_none.h"
#include "dec_nuke.h"
#include "dec_pp.h"
#include "dec_rake.h"
#include "dec_rdcn.h"
#include "dec_rlen.h"
//#include "dec_rncd.h"
#include "dec_sdhc.h"
#include "dec_shr3.h"
#include "dec_shri.h"
#include "dec_slz3.h"
#include "dec_smpl.h"
#include "dec_sqsh.h"
#include "dec_sxsc.h"
#include "dec_tdcs.h"
//#include "dec_tpwm.h"
#include "dec_zeno.h"
#include "xpk.h"

XPKDecompressor *
xpk_lookup_decompressor(xpk_compression_t comp,const Buffer &buf,std::unique_ptr<XPKDecompressor::State> &state)
{
    switch(comp){
    case XPK_ACCA: return new ACCADecompressor(buf);
    case XPK_ARTM: return new ARTMDecompressor(buf);
    case XPK_BLZW: return new BLZWDecompressor(buf);
    case XPK_CBR0: return new CBR0Decompressor(buf, false);
    case XPK_CBR1: return new CBR0Decompressor(buf, true);
    case XPK_CRMS: return new CRMDecompressor(buf, true);
    case XPK_CRM2: return new CRMDecompressor(buf, false);
    case XPK_CYB2:
        // FIXME!! required recursive compression to work
        return NULL;
    case XPK_DUKE: return new NUKEDecompressor(buf, true);
    case XPK_ELZX: return new LZXDecompressor(buf, false, false);
    case XPK_FAST: return new FASTDecompressor(buf);
    case XPK_FBR2: return new FBR2Decompressor(buf);
    case XPK_FRHT: return new RAKEDecompressor(buf, false);
    case XPK_FRLE: return new FRLEDecompressor(buf);
    case XPK_HFMN: return new HFMNDecompressor(buf);
    case XPK_HUFF: return new HUFFDecompressor(buf);
    case XPK_ILZR: return new ILZRDecompressor(buf);
    case XPK_IMP:  return new IMPDecompressor(buf);
    case XPK_LHLB: return new LHLBDecompressor(buf);
    case XPK_LIN1: return new LIN1Decompressor(buf, false);
    case XPK_LIN2: return new LIN2Decompressor(buf, false);
    case XPK_LIN3: return new LIN1Decompressor(buf, true);
    case XPK_LIN4: return new LIN2Decompressor(buf, true);
    case XPK_LZBS: return new LZBSDecompressor(buf);
    case XPK_LZCB: return new LZCBDecompressor(buf);
    case XPK_LZW2: return new LZW2Decompressor(buf, 2);
    case XPK_LZW3: return new LZW2Decompressor(buf, 3);
    case XPK_LZW4: return new LZW4Decompressor(buf);
    case XPK_LZW5: return new LZW5Decompressor(buf);
    case XPK_MASH: return new MASHDecompressor(buf);
    case XPK_NONE: return new NONEDecompressor(buf);
    case XPK_NUKE: return new NUKEDecompressor(buf, false);
    case XPK_PWPK: return new PPDecompressor(buf, state);
    case XPK_RAKE: return new RAKEDecompressor(buf, true);
    case XPK_RDCN: return new RDCNDecompressor(buf);
    case XPK_RLEN: return new RLENDecompressor(buf);
    case XPK_SASC: return new SXSCDecompressor(buf, false);
    case XPK_SDHC:
        // FIXME!! required recursive compression to work
        //return new SDHCDecompressor(src_buf, state, false);
        return NULL;
    case XPK_SHR3: return new SHR3Decompressor(buf, state);
    case XPK_SHRI: return new SHRIDecompressor(buf, state);
    case XPK_SHSC: return new SXSCDecompressor(buf, true);
    case XPK_SLZX: return new LZXDecompressor(buf, false, true);
    case XPK_SLZ3: return new SLZ3Decompressor(buf);
    case XPK_SMPL: return new SMPLDecompressor(buf);
    case XPK_SQSH: return new SQSHDecompressor(buf);
    case XPK_TDCS: return new TDCSDecompressor(buf);
    case XPK_ZENO: return new ZENODecompressor(buf);
    }
    return NULL;
}
