/*
 * Copyright (C) 2020 - 2023  Magnus Bergman <magnus@snisurset.net>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */
#include "loa-buffer.h"


loa_buffer_t *
loa_buffer_init(loa_buffer_t *b, int object_size, int layer_size, int anim_size, int keyframe_size)
{
    loa_array_init(&b->point_array, sizeof(loa_point_t));
    loa_array_init(&b->segment_array, sizeof(loa_segment_t));
    loa_array_init(&b->object_array, object_size);
    loa_array_init(&b->layer_array, layer_size);
    loa_array_init(&b->anim_array, anim_size);
    loa_array_init(&b->keyframe_array, keyframe_size);
    return b;
}


void
loa_buffer_done(loa_buffer_t *b)
{
    loa_array_done(&b->point_array);
    loa_array_done(&b->segment_array);
    loa_array_done(&b->object_array);
    loa_array_done(&b->layer_array);
    loa_array_done(&b->anim_array);
    loa_array_done(&b->keyframe_array);
}


void
loa_buffer_clear(loa_buffer_t *b)
{
    loa_array_clear(&b->point_array);
    loa_array_clear(&b->segment_array);
    loa_array_clear(&b->object_array);
    loa_array_clear(&b->layer_array);
    loa_array_clear(&b->keyframe_array);
    loa_array_clear(&b->anim_array);
}


loa_point_t *
loa_buffer_push_segment(loa_buffer_t *b, loa_path_t *path, loa_segment_type_t type)
{
    loa_segment_t *seg = loa_buffer_add_segment(b);
    int points;
    seg->type = type;
    seg->point = loa_buffer_point_count(b);
    ++path->segment_count;
    switch(type){
    case LOA_GOTO:
    case LOA_LINE:
        points = 1;
        break;
    case LOA_CURVE:
        points = 3;
        break;
    case LOA_CLOSE:
        return NULL;
    }
    path->point_count += points;
    return loa_buffer_add_points(b, points);
}


loa_path_t *
loa_buffer_begin_path(loa_buffer_t *b, loa_path_t *path)
{
    path->segment = loa_buffer_segment_count(b);
    return path;
}


void loa_buffer_end_path(loa_buffer_t *b, loa_path_t *path)
{
    loa_segment_t *seg = loa_buffer_get_segment(b, path->segment);
    path->segment_count = loa_buffer_segment_count(b) - path->segment;
    path->point_count = loa_buffer_point_count(b) - seg->point;
}
