/*
 * Copyright (C) 2004 - 2023  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <debug.h>
#include <endian.h>
#include <string.h>                     /* memset() */

#include "nil-amiga.h"


#ifndef MIN
# define MIN(a,b) ((a)<(b)?(a):(b))
#endif
#ifndef MAX
# define MAX(a,b) ((a)>(b)?(a):(b))
#endif
#define CLAMP_BYTE(a) MAX(0,MIN(255,a))


#if 0
void
amiga_decode_palette(const uint16_t *amiga_palette, amiga_cairo_color_t *cairo_palette,int colors)
{
    int i;
    for(i=0;i<colors;++i){
        uint16_t col = be16toh(amiga_palette[i]);
        int r = (col>>8)&0x0f;
        int g = (col>>4)&0x0f;
        int b = (col>>0)&0x0f;
#ifdef AMIGA_CAIRO_BITS16
        r = (r<<1) | (r>>3);
        g = (g<<2) | (g>>2);
        b = (b<<1) | (b>>3);
        cairo_palette[i] = (r<<11) | (g<<5) | (b<<0);
#else
        r = (r<<4) | r;
        g = (g<<4) | g;
        b = (b<<4) | b;
        cairo_palette[i] = (r<<16) | (g<<8) | b;
#endif
    }
}
#endif


#if 0
void
abydos_copy_line_from_amiga_bitplanes(const uint8_t **amiga_bitplane, amiga_cairo_color_t *dst, int bplcon0, amiga_cairo_color_t *palette)
{
}
#endif


void
nil_amiga_copy_colors(uint8_t *dst, int bpc, nile_format_t format, const uint16_t *src, int count)
{
    int r_o = NILE_OFFSET_RED(format);
    int g_o = NILE_OFFSET_GREEN(format);
    int b_o = NILE_OFFSET_BLUE(format);
    while(count-->0){
        int col = be16toh(*src++);
        int r = (col>>8)&0x0f;
        int g = (col>>4)&0x0f;
        int b = (col>>0)&0x0f;
        dst[r_o] = r<<4|r;
        dst[g_o] = g<<4|r;
        dst[b_o] = b<<4|r;
        dst += bpc;
    }
}


#if 0
cairo_surface_t *
abydos_image_surface_create_from_bitplanes(const uint8_t **bitplane,
                                           int width, int height, int depth,
                                           const amiga_cairo_color_t *palette, int palette_stepping)
{
    int src_offs = 0;
    int src_stride = nil_amiga_plane_stride(width);
    cairo_surface_t *surface = cairo_image_surface_create(CAIRO_FORMAT_RGB24,width,height);
    uint32_t *dst = (uint32_t*)cairo_image_surface_get_data(surface);
    int dst_stride = cairo_image_surface_get_stride(surface)/sizeof(uint32_t);

    while(height-->0){
        int x;
        for(x=0;x<src_stride;++x){
            int plane;
            uint8_t b[8];
            int i;
            memset(b,0,8);
            for(plane=0;plane<depth;++plane){
                for(i=0;i<8;++i)
                    if(bitplane[plane][src_offs+x] & (128>>i))
                        b[i] |= 1<<plane;
            }
            for(i=0;i<8;++i){
                if(x*8+i >= width)
                    break;
                dst[x*8+i] = palette[b[i]];
            }
        }
        src_offs += src_stride;
        dst += dst_stride;
        palette += palette_stepping;
    }

    cairo_surface_mark_dirty(surface);
    return surface;
}
#endif


#if 0
cairo_surface_t *
abydos_image_surface_create_from_ehb_bitplanes(const uint8_t **bitplane,
                                               int width, int height, int depth,
                                               const amiga_cairo_color_t *palette, int palette_stepping)
{
    int src_offs = 0;
    int src_stride = amiga_plane_stride(width);
    cairo_surface_t *surface = cairo_image_surface_create(CAIRO_FORMAT_RGB24,width,height);
    uint32_t *dst = (uint32_t*)cairo_image_surface_get_data(surface);
    int dst_stride = cairo_image_surface_get_stride(surface)/sizeof(uint32_t);
    int mask = ~(1<<(depth-1));

    while(height-->0){
        uint32_t rgb = 0;
        int x;
        for(x=0;x<src_stride;++x){
            int plane;
            uint8_t b[8];
            int i;
            memset(b,0,8);
            for(plane=0;plane<depth;++plane){
                for(i=0;i<8;++i)
                    if(bitplane[plane][src_offs+x] & (128>>i))
                        b[i] |= 1<<plane;
            }
            for(i=0;i<8;++i){
                int pix;
                if(x*8+i >= width)
                    break;
                pix = b[i];
                if(pix & ~mask){
                    rgb = (palette[pix&mask]>>1) & 0x00777777;
                }else{
                    rgb = palette[pix];
                }
                dst[x*8+i] = rgb;
            }
        }
        src_offs += src_stride;
        dst += dst_stride;
        palette += palette_stepping;
    }

    cairo_surface_mark_dirty(surface);
    return surface;
}
#endif


#if 0
cairo_surface_t *
abydos_image_surface_create_from_ham_bitplanes(const uint8_t **bitplane,

                                               int width, int height, int depth,
                                               const amiga_cairo_color_t *palette, int palette_stepping)
{
    int src_offs = 0;
    int src_stride = amiga_plane_stride(width);
    cairo_surface_t *surface = cairo_image_surface_create(CAIRO_FORMAT_RGB24,width,height);
    uint32_t *dst = (uint32_t*)cairo_image_surface_get_data(surface);
    int dst_stride = cairo_image_surface_get_stride(surface)/sizeof(uint32_t);
    int hold = depth > 6 ? 6 : 4;

    while(height-->0){
        uint32_t rgb = 0;
        int x;
        for(x=0;x<src_stride;++x){
            int plane;
            uint8_t b[8];
            int i;
            memset(b,0,8);
            for(plane=0;plane<depth;++plane){
                for(i=0;i<8;++i)
                    if(bitplane[plane][src_offs+x] & (128>>i))
                        b[i] |= 1<<plane;
            }
            for(i=0;i<8;++i){
                int pix;
                if(x*8+i >= width)
                    break;
                pix = b[i];
                switch(pix>>hold){
                case 0:
                    rgb = palette[pix];
                    break;
                case 1: /* blue */
                    pix = pix << (8-hold) & 0xff;
                    pix |= pix >> hold;
                    rgb = (rgb & 0xffff00) | pix;
                    break;
                case 2: /* red */
                    pix = pix << (8-hold) & 0xff;
                    pix |= pix >> hold;
                    rgb = (rgb & 0x00ffff) | pix << 16;
                    break;
                case 3: /* green */
                    pix = pix << (8-hold) & 0xff;
                    pix |= pix >> hold;
                    rgb = (rgb & 0xff00ff) | pix << 8;
                    break;
                }
                dst[x*8+i] = rgb;
            }
        }
        src_offs += src_stride;
        dst += dst_stride;
        palette += palette_stepping;
    }

    cairo_surface_mark_dirty(surface);
    return surface;
}
#endif


#if 0
cairo_surface_t *
abydos_image_surface_create_from_amiga_bitplanes(const uint8_t **bitplane,
                                                 int width, int height, int depth, int bplcon0,
                                                 const amiga_cairo_color_t *palette, int palette_size, int palette_stepping)
{
    if(1<<depth > palette_size){
        if(bplcon0 & BPLCON0_HAM){
            DEBUG_p("HAM");
            return abydos_image_surface_create_from_ham_bitplanes(bitplane, width, height, depth, palette, palette_stepping);
        }
        if(!(bplcon0 & BPLCON0_DPF) && 6==depth){
            DEBUG_p("EHB");
            return abydos_image_surface_create_from_ehb_bitplanes(bitplane, width, height, depth, palette, palette_stepping);
        }
    }
    return abydos_image_surface_create_from_bitplanes(bitplane, width, height, depth, palette, palette_stepping);
}
#endif


#if 0
cairo_surface_t *
abydos_image_surface_create_from_amiga_pixels(const uint8_t *pixels,
                                              int width, int height,
                                              int stride, int bplcon0,
                                              amiga_cairo_color_t *palette,
                                              int palette_stepping)
{
}
#endif


static void
_blit_from_bitplanes_ehb(nile_blit_t *blit, uint8_t *dst, uint8_t **src, off_t offs, int count)
{
    int mask = 1<<(blit->src_depth-1);
    while(count>0){
        uint8_t b[8];
        int plane;
        int i;
        memset(b,0,8);
        for(plane=0;plane<blit->src_depth;++plane)
            for(i=0;i<8;++i)
                if(src[plane][offs] & (128>>i))
                    b[i] |= 1<<plane;
        for(i=0;i<8;++i){
            int idx;
            const uint8_t *color;
            if(count <= 0)
                return;
            idx = b[i];
            color = blit->palette + (idx & ~mask) * blit->dst_bpc;
            if(idx & mask){
                dst[blit->dst_o[0]] = (color[blit->dst_o[0]]>>1) & 0x77;
                dst[blit->dst_o[1]] = (color[blit->dst_o[1]]>>1) & 0x77;
                dst[blit->dst_o[2]] = (color[blit->dst_o[2]]>>1) & 0x77;
            }else{
                memcpy(dst, color, blit->dst_bpc);
            }
            dst += blit->dst_bpc;
            --count;
        }
        ++offs;
    }
}


static void
_blit_from_bitplanes_ham(nile_blit_t *blit, uint8_t *dst, uint8_t **src, off_t offs, int count)
{
    int hold = blit->src_depth > 6 ? 6 : 4;
    uint8_t color[NILE_MAX_BYTES_PER_COLOR];
    memset(color, 0, blit->dst_bpc);
    while(count>0){
        uint8_t b[8];
        int plane;
        int i;
        memset(b,0,8);
        for(plane=0;plane<blit->src_depth;++plane)
            for(i=0;i<8;++i)
                if(src[plane][offs] & (128>>i))
                    b[i] |= 1<<plane;
        for(i=0;i<8;++i){
            int idx;
            int channel;
            if(count <= 0)
                return;
            idx = b[i];
            channel = idx>>hold;
            if(channel){
                idx = idx << (8-hold) & 0xff;
                idx |= idx >> hold;
                color[blit->dst_o[channel-1]] = idx;
            }else{
                memcpy(color, blit->palette + idx * blit->dst_bpc, blit->dst_bpc);
            }
            memcpy(dst, color, blit->dst_bpc);
            dst += blit->dst_bpc;
            --count;
        }
        ++offs;
    }
}


nile_blit_t *
nil_amiga_blit_init_from_bitplanes(nile_blit_t *blit,
                                   int dst_bpc,
                                   nile_format_t dst_format,
                                   int depth,
                                   const uint8_t *palette,
                                   int colors,
                                   int src_bpc,
                                   nile_format_t src_format,
                                   int bplcon0)
{
    nile_blit_init_from_bitplanes_unindex_rgb(blit, dst_bpc, dst_format, depth, palette, colors, src_bpc, src_format);
    /* Use amiga channel ordering. */
    blit->dst_o[1] = NILE_OFFSET_RED(dst_format);
    blit->dst_o[2] = NILE_OFFSET_GREEN(dst_format);
    blit->dst_o[0] = NILE_OFFSET_BLUE(dst_format);
    if(bplcon0 & BPLCON0_HAM){
        DEBUG_p("HAM");
        blit->blit_from_bitplanes_func = _blit_from_bitplanes_ham;
    }else if(!(bplcon0 & BPLCON0_DPF) && 6==depth){
        DEBUG_p("EHB");
        blit->blit_from_bitplanes_func = _blit_from_bitplanes_ehb;
    }
    return blit;
}


void
nil_amiga_blit_direct_from_bitplanes(uint8_t *dst,
                                     int dst_stride,
                                     int dst_bpc,
                                     nile_format_t dst_format,
                                     uint8_t **src,
                                     int src_stride,
                                     int depth,
                                     const uint8_t *palette,
                                     int colors,
                                     int src_bpc,
                                     nile_format_t src_format,
                                     int width,
                                     int height,
                                     int bplcon0,
                                     int palette_stepping,
                                     int flags)
{
    nile_blit_t blit;
    off_t offs = 0;
    nil_amiga_blit_init_from_bitplanes(&blit, dst_bpc, dst_format, depth, palette, palette_stepping ? height*palette_stepping : colors, src_bpc, src_format, bplcon0);
    while(height-->0){
        nile_blit_from_bitplanes_line(&blit, dst, src, offs, width);
        blit.palette += palette_stepping * src_bpc;
        dst += dst_stride;
        offs += src_stride;
    }
}


#if 0
void nil_amiga_direct_from_ehb_bitplanes(nile_t *ni, int palette_stepping)
{
    int src_offs = 0;
    int src_stride = amiga_plane_stride(width);
    cairo_surface_t *surface = cairo_image_surface_create(CAIRO_FORMAT_RGB24,width,height);
    uint32_t *dst = (uint32_t*)cairo_image_surface_get_data(surface);
    int dst_stride = cairo_image_surface_get_stride(surface)/sizeof(uint32_t);
    int mask = ~(1<<(depth-1));

    while(height-->0){
        uint32_t rgb = 0;
        int x;
        for(x=0;x<src_stride;++x){
            int plane;
            uint8_t b[8];
            int i;
            memset(b,0,8);
            for(plane=0;plane<depth;++plane){
                for(i=0;i<8;++i)
                    if(bitplane[plane][src_offs+x] & (128>>i))
                        b[i] |= 1<<plane;
            }
            for(i=0;i<8;++i){
                int pix;
                if(x*8+i >= width)
                    break;
                pix = b[i];
                if(pix & ~mask){
                    rgb = (palette[pix&mask]>>1) & 0x00777777;
                }else{
                    rgb = palette[pix];
                }
                dst[x*8+i] = rgb;
            }
        }
        src_offs += src_stride;
        dst += dst_stride;
        palette += palette_stepping;
    }

    cairo_surface_mark_dirty(surface);
    return surface;
}
#endif


#if 0
void nil_amiga_planar_from_ham_bitplanes(nile_t *ni, int palette_stepping)
{
    int src_offs = 0;
    int src_stride = amiga_plane_stride(width);
    cairo_surface_t *surface = cairo_image_surface_create(CAIRO_FORMAT_RGB24,width,height);
    uint32_t *dst = (uint32_t*)cairo_image_surface_get_data(surface);
    int dst_stride = cairo_image_surface_get_stride(surface)/sizeof(uint32_t);
    int hold = depth > 6 ? 6 : 4;

    while(height-->0){
        uint32_t rgb = 0;
        int x;
        for(x=0;x<src_stride;++x){
            int plane;
            uint8_t b[8];
            int i;
            memset(b,0,8);
            for(plane=0;plane<depth;++plane){
                for(i=0;i<8;++i)
                    if(bitplane[plane][src_offs+x] & (128>>i))
                        b[i] |= 1<<plane;
            }
            for(i=0;i<8;++i){
                int pix;
                if(x*8+i >= width)
                    break;
                pix = b[i];
                switch(pix>>hold){
                case 0:
                    rgb = palette[pix];
                    break;
                case 1: /* blue */
                    pix = pix << (8-hold) & 0xff;
                    pix |= pix >> hold;
                    rgb = (rgb & 0xffff00) | pix;
                    break;
                case 2: /* red */
                    pix = pix << (8-hold) & 0xff;
                    pix |= pix >> hold;
                    rgb = (rgb & 0x00ffff) | pix << 16;
                    break;
                case 3: /* green */
                    pix = pix << (8-hold) & 0xff;
                    pix |= pix >> hold;
                    rgb = (rgb & 0xff00ff) | pix << 8;
                    break;
                }
                dst[x*8+i] = rgb;
            }
        }
        src_offs += src_stride;
        dst += dst_stride;
        palette += palette_stepping;
    }

    cairo_surface_mark_dirty(surface);
    return surface;
}
#endif


void
nil_amiga_direct_from_bitplanes(nile_t *dst, const nile_t *src, uint32_t bplcon0, int palette_stepping, int flags)
{
#if 1
    nile_ensure_direct(dst, src->width, src->height, src->bpc);
    if(!dst->format)
        dst->format = src->format;
    nil_amiga_blit_direct_from_bitplanes(dst->pixels, nile_get_stride(dst),
                                         dst->bpc, dst->format,
                                         src->bitplanes, nile_get_stride(src), src->depth,
                                         src->palette, src->colors,
                                         src->bpc, src->format,
                                         src->width, src->height,
                                         bplcon0, palette_stepping, 0);
#else
    if(1<<ni->depth > ni->colors){
        if(bplcon0 & BPLCON0_HAM){
            DEBUG_p("HAM");
            return abydos_image_surface_create_from_ham_bitplanes(bitplane, width, height, depth, palette, palette_stepping);
        }
        if(!(bplcon0 & BPLCON0_DPF) && 6==depth){
            DEBUG_p("EHB");
            return abydos_image_surface_create_from_ehb_bitplanes(bitplane, width, height, depth, palette, palette_stepping);
        }
    }
#endif
}


static int
_get_dctv_value(uint8_t **bitplane, int depth, int offs, int x, const int *table){
    int mask;
    int plane;
    int c = 0;
    offs += x/8;
    mask = 128>>(x&7);
    for(plane=0;plane<depth;++plane)
        if(bitplane[plane][offs]&mask)
            c |= 1<<plane;
    return table[c];
}


int
_is_dctv_sub(const nile_t *ni, int offs, const int *table){
    int r;
    int x;
    if(_get_dctv_value(ni->bitplanes, ni->depth, offs, 0, table) >> 6 != 0)
        return 0;
    r = 0x7d;
    for(x=1;x<256;++x){
        if(_get_dctv_value(ni->bitplanes, ni->depth, offs, x, table) >> 6 == (r & 1))
            return 0;
        if ((r & 1) != 0)
        r ^= 0xc3 << 1;
        r >>= 1;
    }
    return 1;
}


static void
_make_dctv_table(int *table, uint8_t *palette, int colors, int bpc, nile_format_t format)
{
    int r_o = NILE_OFFSET_RED(format);
    int g_o = NILE_OFFSET_GREEN(format);
    int b_o = NILE_OFFSET_BLUE(format);
    int i;
    for(i=0;i<colors;++i){
        int r = palette[i * bpc + r_o];
        int g = palette[i * bpc + g_o];
        int b = palette[i * bpc + b_o];
        /* The value is 0I0R0B0G.
         * Digital R, G, B are the highest bits of the components.
         * I is the lowest bit of the 4-bit Blue component.
         */
        table[i] = (b << 2 & 0x40) | (r >> 3 & 0x10) | (b >> 5 & 4) | (g >> 7 & 1);
    }
}


int
nil_amiga_is_dctv(const nile_t *ni, uint32_t bplcon0){
    int table[256];
    if(bplcon0 & BPLCON0_HAM)
        return 0;
    _make_dctv_table(table, ni->palette, ni->colors, ni->bpc, ni->format);
    if(!_is_dctv_sub(ni, 0, table))
        return 0;
    if(bplcon0 & BPLCON0_LACE)
        if(!_is_dctv_sub(ni, nil_amiga_plane_stride(ni->width), table))
            return 0;
    return 1;
}


int
nil_amiga_decode_dctv(nile_t *dst_ni, const nile_t *src, uint32_t bplcon0)
{
    int table[256];
    int lace = (bplcon0 & BPLCON0_LACE) != 0;
    int skiprows = lace ? 2 : 1;
    int r_o;
    int g_o;
    int b_o;
    int src_stride;
    int src_offs;
    uint8_t *dst;
    int dst_skip;
    int chroma[2560];
    int y = 0;

    _make_dctv_table(table, src->palette, src->colors, src->bpc, src->format);

    src_stride = nil_amiga_plane_stride(src->width);
    nile_ensure_direct(dst_ni, src->width, src->height, 3);
    if(!dst_ni->format)
        dst_ni->format = src->format;
    r_o = NILE_OFFSET_RED(dst_ni->format);
    g_o = NILE_OFFSET_GREEN(dst_ni->format);
    b_o = NILE_OFFSET_BLUE(dst_ni->format);
    dst = dst_ni->pixels;
    dst_skip = nile_get_skip(dst_ni);
    {
        int dst_stride = nile_get_stride(dst_ni);
        memset(dst, 0, skiprows * dst_stride);
        dst += skiprows * dst_stride;
        src_offs = skiprows * src_stride;
        //dst_ni->height -= skiprows;
    }

    for(;y<src->height-skiprows;++y){
        int odd = y >> lace & 1;
        int r = 0;
        int g = 0;
        int b = 0;
        int o = 0;
        int p = 0;
        int x;
        for(x=0;x<src->width;++x){
            if ((x & 1) == odd){
                int n = x + 1 < src->width ? _get_dctv_value(src->bitplanes, src->depth, src_offs, x, table) << 1 | _get_dctv_value(src->bitplanes, src->depth, src_offs, x+1, table) : 0;
                int i = (o + n) >> 1;
                i = i <= 64 ? 0 : i >= 224 ? 255 : (i - 64) * 8 / 5;

                int u = n + p - (o << 1);
                if (u < 0)
                    u += 3;
                u >>= 2;
                if (((x + 1) & 2) == 0)
                    u = -u;
                int chroma_offs = (x & ~1) | (y & lace);
                int v = y > lace ? chroma[chroma_offs] : 0;
                chroma[chroma_offs] = u;
                if(0 == odd){
                    u = v;
                    v = chroma[chroma_offs];
                }

                p = o;
                o = n;

                r = i + v * 4655 / 2560;
                b = i + u * 8286 / 2560;
                g = i - (v * 2372 + u * 1616) / 2560;
                //rgb = CLAMP_BYTE(r) << 16 | CLAMP_BYTE(g) << 8 | CLAMP_BYTE(b);
                r = CLAMP_BYTE(r);
                g = CLAMP_BYTE(g);
                b = CLAMP_BYTE(b);
            }
            dst[r_o] = r;
            dst[g_o] = g;
            dst[b_o] = b;
            dst += dst_ni->bpc;
        }
        dst += dst_skip;
        src_offs += src_stride;
    }
    return 0;
}


static int
_get_hame_byte(uint8_t **bitplane, int offs){
    int himask = 128>>((offs&3)<<1);
    int lomask = 64>>((offs&3)<<1);
    int plane;
    int c = 0;
    for(plane=0;plane<4;++plane){
        int bits = bitplane[plane][offs>>2];
        if(bits & himask)
            c |= 16<<plane;
        if(bits & lomask)
            c |= 1<<plane;
    }
    return c;
}


int
nil_amiga_is_hame(const nile_t *ni, int offs){
    int i;
    for(i=0;i<7;++i){
        static const uint8_t magic[7] = {0xa2, 0xf5, 0x84, 0xdc, 0x6d, 0xb0, 0x7f};
        if(_get_hame_byte(ni->bitplanes, offs+i) != magic[i])
            return 0;
    }
    switch(_get_hame_byte(ni->bitplanes, offs+7)){
    case 0x14: // REG
    case 0x18: // HAM-E
        return 1;
    default:
        return 0;
    }
}


void
nil_amiga_direct_from_hame(nile_t *dst_ni, const nile_t *src, uint32_t bplcon0){
    int lace = (bplcon0 & BPLCON0_LACE);
    int src_offs = 0;
    uint8_t *dst;
    int dst_skip;
    int r_o;
    int g_o;
    int b_o;
    uint8_t palette[NILE_MAX_BYTES_PER_COLOR * 512];
    int palette_length[2] = {0,0};
    int hame = 0;
    int y;

    nile_ensure_direct(dst_ni, src->width, src->height, 3);
    if(!dst_ni->format)
        dst_ni->format = NILE_FORMAT_r8g8b8;
    dst = dst_ni->pixels;
    dst_skip = nile_get_skip(dst_ni);
    r_o = NILE_OFFSET_RED(dst_ni->format);
    g_o = NILE_OFFSET_GREEN(dst_ni->format);
    b_o = NILE_OFFSET_BLUE(dst_ni->format);

    memset(palette,0,sizeof(palette));

    for(y=0;y<src->height;++y){
        int palette_offs = lace && y&1 ? 0 : 256;
        if(nil_amiga_is_hame(src, src_offs)){
            int c;
            palette_offs += palette_length[palette_offs >> 8];
            DEBUG_p("%04d: palette line",y);
            for(c=0;c<64;++c){
                int r = _get_hame_byte(src->bitplanes, src_offs + 8+c*3 + 0);
                int g = _get_hame_byte(src->bitplanes, src_offs + 8+c*3 + 1);
                int b = _get_hame_byte(src->bitplanes, src_offs + 8+c*3 + 2);
                palette[(palette_offs + c) * dst_ni->bpc + r_o] =  r;
                palette[(palette_offs + c) * dst_ni->bpc + g_o] =  g;
                palette[(palette_offs + c) * dst_ni->bpc + b_o] =  b;
            }
            palette_length[palette_offs >> 8] = (palette_length[palette_offs >> 8] + 64) & 0xff;
            hame = _get_hame_byte(src->bitplanes,src_offs+7) == 0x18;
            memset(dst, 0, src->width * dst_ni->bpc); /* blank the special line */
            dst += src->width * dst_ni->bpc;
            src_offs += src->width;
        }else{
            int palette_bank = 0;
            int r = 0;
            int g = 0;
            int b = 0;
            int x;
            if(hame){
                DEBUG_p("%04d: HAM-E line",y);
                for(x=0;x<src->width;++x){
                    int c = _get_hame_byte(src->bitplanes,src_offs++);
                    switch(c>>6){
                    case 2: /* red */
                        r = (c & 0x3f)<<2;
                        r |= r>>6;
                        break;
                    case 3: /* green */
                        g = (c & 0x3f)<<2;
                        g |= r>>6;
                        break;
                    case 1: /* blue */
                        b = (c & 0x3f)<<2;
                        b |= r>>6;
                        break;
                    default:
                        if (c < 60){
                            const uint8_t *color = &palette[(palette_offs + palette_bank + c) * dst_ni->bpc];
                            r = color[r_o];
                            g = color[g_o];
                            b = color[b_o];
                        }else{
                            palette_bank = (c - 60) << 6;
                        }
                        break;
                    }
                    dst[r_o] = r;
                    dst[g_o] = g;
                    dst[b_o] = b;
                    dst += dst_ni->bpc;
                }
            }else{
                DEBUG_p("%04d: normal line line",y);
                for(x=0;x<src->width;++x){
                    memcpy(dst, palette + (palette_offs + _get_hame_byte(src->bitplanes, src_offs++)) * dst_ni->bpc, dst_ni->bpc);
                    dst += dst_ni->bpc;
                }
            }
        }
        dst += dst_skip;
    }
}
