/*
 * Copyright (C) 2023  Magnus Bergman <magnus@snisurset.net>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */
#include <string.h>                     /* memcpy() */

#include "nil-packed.h"


static void
_blit_packed_unindex(nile_blit_t *blit, uint8_t *dst, const uint8_t *src, int count)
{
    int ppb = 8 / blit->src_depth; /* Pixels per byte */
    int mask = 0xff >> (8-blit->src_depth);
    int i;
    for(i=0;i<count;++i){
        int pix;
        if(0 == i % ppb)
            pix = *src++;
        pix <<= blit->src_depth;
        memcpy(dst, blit->palette + ((pix>>8) & mask) * blit->dst_bpc, blit->dst_bpc);
        dst += blit->dst_bpc;
    }
}


nile_blit_t *
nil_packed_blit_init_unindex_rgb(nile_blit_t *blit,
                                 int dst_bpc,
                                 nile_format_t dst_format,
                                 int depth,
                                 const uint8_t *palette,
                                 int colors,
                                 int src_bpc,
                                 nile_format_t src_format)
{
    if(nile_blit_setup_palette(blit, dst_bpc, dst_format, palette, colors, src_bpc, src_format) < 0)
        return NULL;
    blit->dst_bpc = dst_bpc;
    blit->src_depth = depth;
    blit->blit_func = _blit_packed_unindex;
    return blit;
}
