/*
 * Copyright (C) 2023  Magnus Bergman <magnus@snisurset.net>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */
#include "debug.h"
#include "nile-blit-comp.h"


static void
_blit_8i3(nile_blit_t *blit, uint8_t *dst, uint8_t *src[4], off_t offs, int count)
{
    while(count-->0){
        dst[blit->dst_o[0]] = src[blit->src_o[0]][offs];
        dst[blit->dst_o[1]] = src[blit->src_o[1]][offs];
        dst[blit->dst_o[2]] = src[blit->src_o[2]][offs];
        dst += blit->dst_bpc;
        offs += blit->src_bpc;
    }
}


static void
_blit_8i3a(nile_blit_t *blit, uint8_t *dst, uint8_t *src[4], off_t offs, int count)
{
    while(count-->0){
        dst[blit->dst_o[0]] = src[blit->src_o[0]][offs];
        dst[blit->dst_o[1]] = src[blit->src_o[1]][offs];
        dst[blit->dst_o[2]] = src[blit->src_o[2]][offs];
        dst[blit->dst_o[4]] = src[blit->src_o[4]][offs];
        dst += blit->dst_bpc;
        offs += blit->src_bpc;
    }
}


static void
_blit_8i3a_multiply(nile_blit_t *blit, uint8_t *dst, uint8_t *src[4], off_t offs, int count)
{
    while(count-->0){
        int a = src[blit->src_o[4]][offs];
        dst[blit->dst_o[0]] = a * src[blit->src_o[0]][offs] / 255;
        dst[blit->dst_o[1]] = a * src[blit->src_o[1]][offs] / 255;
        dst[blit->dst_o[2]] = a * src[blit->src_o[2]][offs] / 255;
        dst += blit->dst_bpc;
        offs += blit->src_bpc;
    }
}


#if 0
static void
_blit_g16(uint8_t *dstb, uint8_t *src[4], size_t offs, int count)
{
    uint32_t *dst = (uint32_t*)dstb;
#if CAIRO_VERSION >= CAIRO_VERSION_ENCODE(1,17,2)
    float *dstf = (float*)dst;
    while(count-->0){
        float g = (float)*(uint16_t*)&src[0][offs] / INT16_MAX;
        *dstf++ = g;
        offs += 2;
    }
#elif CAIRO_VERSION >= CAIRO_VERSION_ENCODE(1,12,0)
    while(count-->0){
        int g = *(uint16_t*)&src[0][offs] >> 6;
        *dst++ = g << 20 | g << 10 | g << 0;
        offs += 2;
    }
#else
    while(count-->0){
        int g = *(uint16_t*)&src[0][offs] >> 8;
        *dst++ = g << 16 | g << 8 | g << 0;
        offs += 2;
    }
#endif
}




static void
_blit_ga16(uint8_t *dst, uint8_t *src[4], size_t offs, int count)
{
#if CAIRO_VERSION >= CAIRO_VERSION_ENCODE(1,17,2)
    float *dstf = (float*)dst;
    while(count-->0){
        float a = (float)*(uint16_t*)&src[1][offs] / UINT16_MAX;
        float g = a * *(uint16_t*)&src[0][offs] / UINT16_MAX;
        *dstf++ = g;
        *dstf++ = a;
        offs += 2;
    }
#else
    while(count-->0){
        int a = *(uint16_t*)&src[1][offs] >> 1;
        int g = (a * *(uint16_t*)&src[0][offs] / UINT16_MAX) >> 7;
        *dst++ = g << 16 | g << 8 | g << 0 | (a>>7) << 24;
        offs += 2;
    }
#endif
}


static void
_blit_rgb8(uint8_t *dstb, uint8_t *src[4], size_t offs, int count)
{
    uint32_t *dst = (uint32_t*)dstb;
    while(count-->0){
        int r = src[0][offs];
        int g = src[1][offs];
        int b = src[2][offs];
        *dst++ = r << 16 | g << 8 | b << 0;
        ++offs;
    }
}


static void
_blit_rgb16(uint8_t *dst, uint8_t *src[4], size_t offs, int count)
{
#if CAIRO_VERSION >= CAIRO_VERSION_ENCODE(1,17,2)
    float *dstf = (float*)dst;
    while(count-->0){
        float r = *(uint16_t*)&src[0][offs] / UINT16_MAX;
        float g = *(uint16_t*)&src[1][offs] / UINT16_MAX;
        float b = *(uint16_t*)&src[2][offs] / UINT16_MAX;
        *dstf++ = r;
        *dstf++ = g;
        *dstf++ = b;
        offs += 2;
    }
#elif CAIRO_VERSION >= CAIRO_VERSION_ENCODE(1,12,0)
    while(count-->0){
        int r = *(uint16_t*)&src[0][offs] >> 6;
        int g = *(uint16_t*)&src[1][offs] >> 6;
        int b = *(uint16_t*)&src[2][offs] >> 6;
        *dst++ = r << 20 | g << 10 | b << 0;
        offs += 2;
    }
#else
    while(count-->0){
        int r = *(uint16_t*)&src[0][offs] >> 8;
        int g = *(uint16_t*)&src[1][offs] >> 8;
        int b = *(uint16_t*)&src[2][offs] >> 8;
        *dst++ = r << 16 | g << 8 | b << 0;
        offs += 2;
    }
#endif
}


static void
_blit_rgba8(uint8_t *dstb, uint8_t *src[4], size_t offs, int count)
{
    uint32_t *dst = (uint32_t*)dstb;
    while(count-->0){
        int a = *src[3]++;
        int r = a * src[0][offs] / 255;
        int g = a * src[1][offs] / 255;
        int b = a * src[2][offs] / 255;
        *dst++ = r << 16 | g << 8 | b << 0 | a << 24;
        ++offs;
    }
}


static void
_blit_rgba16(uint8_t *dstb, uint8_t *src[4], size_t offs, int count)
{
#if CAIRO_VERSION >= CAIRO_VERSION_ENCODE(1,17,2)
    float *dst = (float*)dstb;
    while(count-->0){
        float a = (float)*(uint16_t*)&src[3][offs] / UINT16_MAX;
        float r = a * *(uint16_t*)&src[0][offs] / UINT16_MAX;
        float g = a * *(uint16_t*)&src[1][offs] / UINT16_MAX;
        float b = a * *(uint16_t*)&src[2][offs] / UINT16_MAX;
        *dst++ = r;
        *dst++ = g;
        *dst++ = b;
        *dst++ = a;
        offs += 2;
    }
#else
    while(count-->0){
        int a = *(uint16_t*)&src[3][offs] >> 1;
        int r = (a * *(uint16_t*)&src[0][offs] / UINT16_MAX) >> 7;
        int g = (a * *(uint16_t*)&src[1][offs] / UINT16_MAX) >> 7;
        int b = (a * *(uint16_t*)&src[2][offs] / UINT16_MAX) >> 7;
        *dst++ = r << 16 | g << 8 | b << 0 | (a>>7) << 24;
        offs += 2;
    }
#endif
}
#endif


static void
_blit_32f3a_from_16i3a_multiply(nile_blit_t *blit, uint8_t *dst, uint8_t *src[4], off_t offs, int count)
{
    while(count-->0){
        float a = (float)*(uint16_t*)&src[blit->src_o[4]][offs] / UINT16_MAX;
        *(float*)&dst[blit->dst_o[0]] = a * *(uint16_t*)&src[blit->src_o[0]][offs] / UINT16_MAX;
        *(float*)&dst[blit->dst_o[1]] = a * *(uint16_t*)&src[blit->src_o[1]][offs] / UINT16_MAX;
        *(float*)&dst[blit->dst_o[2]] = a * *(uint16_t*)&src[blit->src_o[2]][offs] / UINT16_MAX;
        *(float*)&dst[blit->dst_o[4]] = a;
        dst += blit->dst_bpc;
        offs += blit->src_bpc;
    }
}


nile_blit_t *
nile_blit_init_from_components_rgb(nile_blit_t *blit,
                                   int dst_bpc,
                                   nile_format_t dst_format,
                                   int src_bpc,
                                   nile_format_t src_format)
{
    int comps = 0;
    int i;
    for(i=0; i<5; ++i)
        if(NILE_OFFSET_CHANNEL(src_format, i) < 15)
            ++comps;
    DEBUG_p("%d components", comps);

    blit->dst_bpc = dst_bpc;
    blit->src_bpc = src_bpc;

    blit->dst_o[0] = NILE_OFFSET_CHANNEL(dst_format, 0);
    blit->dst_o[1] = NILE_OFFSET_CHANNEL(dst_format, 1);
    blit->dst_o[2] = NILE_OFFSET_CHANNEL(dst_format, 2);
    blit->dst_o[4] = NILE_OFFSET_CHANNEL(dst_format, 4);
    blit->src_o[0] = NILE_GET_CHANNEL(src_format, 0);
    blit->src_o[1] = NILE_GET_CHANNEL(src_format, 1);
    blit->src_o[2] = NILE_GET_CHANNEL(src_format, 2);
    blit->src_o[4] = NILE_GET_CHANNEL(src_format, 4);
    switch(NILE_GET_BYTES_PER_CHANNEL(dst_format)){
    case 1:
        switch(NILE_GET_BYTES_PER_CHANNEL(src_format)){
        case 1:
            if(NILE_HAS_ALPHA(dst_format) && NILE_HAS_ALPHA(src_format)){
                /* Both have alpha channel */
                if((dst_format & NILE_PREMULTIPLIED) == (src_format & NILE_PREMULTIPLIED)){
                    blit->blit_from_components_func = _blit_8i3a;
                }else if(dst_format & NILE_PREMULTIPLIED){
                    blit->blit_from_components_func = _blit_8i3a_multiply;
                }else{
                    return NULL;
                }
            }else if(NILE_HAS_ALPHA(dst_format)){
                return NULL;
            }else{
                blit->blit_from_components_func = _blit_8i3;
            }
        default:
            return NULL;
        }
        break;
    case 4: /* assume float */
        switch(NILE_GET_BYTES_PER_CHANNEL(src_format)){
        case 2:
            if(NILE_HAS_ALPHA(dst_format) && NILE_HAS_ALPHA(src_format)){
                /* Both have alpha channel */
                if((dst_format & NILE_PREMULTIPLIED) == (src_format & NILE_PREMULTIPLIED)){
                    return NULL;
                }else if(dst_format & NILE_PREMULTIPLIED){
                    blit->blit_from_components_func = _blit_32f3a_from_16i3a_multiply;
                }else{
                    return NULL;
                }
            }else if(NILE_HAS_ALPHA(dst_format)){
                return NULL;
            }else{
                //blit->blit_func_components
            }
            break;
        }
        break;
    default:
        return NULL;
    }
    return blit;
}


void
nile_blit_from_components(nile_blit_t *blit,
                          uint8_t *dst,
                          int dst_stride,
                          uint8_t **src,
                          int src_stride,
                          int width,
                          int height,
                          int flags)
{
    off_t offs = 0;
    if(flags & NILE_FLIPPED){
        dst += dst_stride * (height-1);
        dst_stride *= -1;
    }
    while(height-->0){
        //DEBUG_p("height: %d", height);
        nile_blit_from_bitplanes_line(blit, dst, src, offs, width);
        dst += dst_stride;
        offs += src_stride;
    }
}
