/*
 * Author: Magnus Bergman <magnus@snisurset.net>
 *
 * This file is public domain.
 */
#ifndef nile_blit_h
#define nile_blit_h

#include <nile.h>

typedef struct _nile_blit_t nile_blit_t;
typedef void(*nile_blit_func_t)(nile_blit_t*, uint8_t *dst, const uint8_t *src, int count);
typedef void(*nile_blit_bitplanes_func_t)(nile_blit_t*, uint8_t **dst, off_t, uint8_t **src, off_t, int count);
typedef void(*nile_blit_from_bitplanes_func_t)(nile_blit_t*, uint8_t *dst, uint8_t **src, off_t, int count);
typedef void(*nile_blit_to_bitplanes_func_t)(nile_blit_t*, uint8_t **dst, off_t, const uint8_t *src, int count);
typedef void(*nile_blit_from_components_func_t)(nile_blit_t*, uint8_t *dst, uint8_t **src, off_t, int count);
struct _nile_blit_t{
    union{
        nile_blit_func_t blit_func;
        nile_blit_bitplanes_func_t blit_bitplanes_func;
        nile_blit_from_bitplanes_func_t blit_from_bitplanes_func;
        nile_blit_to_bitplanes_func_t blit_to_bitplanes_func;
        nile_blit_from_components_func_t blit_from_components_func;
    };
    uint8_t tmp_palette[NILE_MAX_BYTES_PER_COLOR * 256];
    const uint8_t *palette;
    union{
        int dst_bpc;
        int dst_depth;
        int dst_comps;
    };
    union{
        int src_bpc;
        int src_depth;
        int src_comps;
    };
    int dst_o[5];
    int src_o[5];
    void *userdata;
};

int nile_blit_setup_palette(nile_blit_t*, int dst_bpc, nile_format_t dst_format, const uint8_t *palette, int colors, int src_bpc, nile_format_t src_format);
nile_blit_t *nile_blit_init_rgb(nile_blit_t*, int dst_bpc, nile_format_t dst_format, int src_bpc, nile_format_t src_format);
nile_blit_t *nile_blit_init_unindex_rgb(nile_blit_t*, int dst_bpc, nile_format_t dst_format, const uint8_t *palette, int colors, int src_bpc, nile_format_t src_format);
nile_blit_t *nile_blit_init_from_bitplanes_rgb(nile_blit_t*, int dst_bpc, nile_format_t dst_format, int depth, nile_format_t src_format);
nile_blit_t *nile_blit_init_from_bitplanes_unindex_rgb(nile_blit_t*, int dst_bpc, nile_format_t dst_format, int depth, const uint8_t *palette, int colors, int src_bpc, nile_format_t src_format);

#define nile_blit_line(a,b,c,d) (a)->blit_func(a,b,c,d)
#define nile_blit_from_bitplanes_line(a,b,c,d,e) (a)->blit_from_bitplanes_func(a,b,c,d,e);
void nile_blit(nile_blit_t*, uint8_t *dst, int dst_stride, const uint8_t *src, int src_stride, int width, int height, int flags);
void nile_blit_from_bitplanes(nile_blit_t*, uint8_t *dst, int dst_stride, uint8_t **src, int src_stride, int width, int height, int flags);

void nile_blit_colors_rgb(uint8_t *dst, int dst_bpc, nile_format_t dst_format, const uint8_t *src, int src_bpc, nile_format_t src_format, int count);

/* Functions to copy pixels between image types without changing the color
 * format. */

void nile_blit_pixels(uint8_t *dst, int dst_stride, const uint8_t *src, int src_stride, int bpc, int width, int height, int flags);
void nile_blit_pixels_unindex(uint8_t *dst, int dst_stride, const uint8_t *src, int src_stride, int bpc, const uint8_t *palette, int width, int height);
void nile_blit_pixels_from_bitplanes(uint8_t *dst, int dst_stride, int dst_pbc, uint8_t **src, int src_stride, int depth, int width, int height);
void nile_blit_pixels_from_bitplanes_unindex(uint8_t *dst, int dst_stride, int bpc, uint8_t **src, int src_stride, int depth, const uint8_t *palette, int width, int height);
void nile_blit_pixels_to_bitplanes(uint8_t **dst, int dst_stride, int depth, const uint8_t *src, int src_stride, int width, int height);

#endif /* nile_blit_h */
