/*
 * Copyright (C) 2023  Magnus Bergman <magnus@snisurset.net>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */
#include "nile.h"
#include "nile-blit.h"


void
nile_convert(nile_t *dst, const nile_t *src, int flags)
{
    nile_blit_t blit;
    switch(dst->type){
    case NILE_TYPE_NONE:
        return;
    case NILE_TYPE_INDEXED:
        nile_ensure_indexed(dst, src->width, src->height);
        switch(src->type){
        case NILE_TYPE_NONE:
            return;
        case NILE_TYPE_INDEXED:
            nile_blit_init_rgb(&blit, 1, NILE_FORMAT_UNKNOWN, 1, NILE_FORMAT_UNKNOWN);
            nile_blit(&blit, dst->pixels, nile_get_stride(dst),
                      src->pixels, nile_get_stride(src), src->width, src->height, flags);
            break;
        case NILE_TYPE_BITPLANES:
            nile_blit_pixels_from_bitplanes(dst->pixels, 1, nile_get_stride(dst),
                                            src->bitplanes, nile_get_stride(src), src->depth, src->width, src->height);
            break;
        case NILE_TYPE_DIRECT:
        case NILE_TYPE_BITPLANES_DIRECT:
            /* not supported, requires color reduction */
            return;
        }
        break;
    case NILE_TYPE_DIRECT:
        nile_ensure_direct(dst, src->width, src->height, dst->bpc);
        switch(src->type){
        case NILE_TYPE_NONE:
            return;
        case NILE_TYPE_INDEXED:
            //nile_copy_colors_rgb(blit.tmp_palette, dst->pbc, dst->format, src->palette,
            nile_blit_init_unindex_rgb(&blit, dst->bpc, dst->format, src->palette, src->colors, src->bpc, src->format);
            nile_blit(&blit, dst->pixels, nile_get_stride(dst),
                      src->pixels, nile_get_stride(src), src->width, src->height, flags);
            break;
        case NILE_TYPE_DIRECT:
            nile_blit_init_rgb(&blit, dst->bpc, dst->format, src->bpc, src->format);
            nile_blit(&blit, dst->pixels, nile_get_stride(dst),
                      src->pixels, nile_get_stride(src), src->width, src->height, flags);
            break;
        case NILE_TYPE_BITPLANES:
            nile_blit_init_from_bitplanes_unindex_rgb(&blit, dst->bpc, dst->format,
                                                      src->depth, src->palette, src->colors, src->bpc, src->format);
            nile_blit_from_bitplanes(&blit, dst->pixels, nile_get_stride(dst),
                                     src->bitplanes, nile_get_stride(src), src->width, src->height, flags);
            break;
        case NILE_TYPE_BITPLANES_DIRECT:
            if(dst->format != src->format){
                /* do a two pass convertion */
                nile_t tmp;
                nile_init(&tmp);
                //nile_copy_identity(&tmp, src);
                //tmp.bpc = src->bpc;
                //tmp.bpc = (src->depth + 7)>>3;
                //tmp.stride = tmp.width * tmp.bpc;
                nile_ensure_direct(&tmp, src->width, src->height, (src->depth+7)>>3);
                tmp.format = src->format;
                nile_blit_pixels_from_bitplanes(tmp.pixels, tmp.stride,
                                                tmp.bpc, src->bitplanes, src->stride, src->depth, src->width, src->height);
                nile_convert(dst, &tmp, flags);
                nile_done(&tmp);
            }else{
                nile_blit_pixels_from_bitplanes(dst->pixels, nile_get_stride(dst), dst->bpc,
                                                src->bitplanes, nile_get_stride(src), src->depth, src->width, src->height);
            }
            break;
        }
        break;
    case NILE_TYPE_BITPLANES:
        switch(src->type){
        case NILE_TYPE_NONE:
            return;
        case NILE_TYPE_INDEXED:
            break;
        case NILE_TYPE_DIRECT:
            break;
        case NILE_TYPE_BITPLANES:
            break;
        case NILE_TYPE_BITPLANES_DIRECT:
            break;
        }
        break;
    case NILE_TYPE_BITPLANES_DIRECT:
        switch(src->type){
        case NILE_TYPE_NONE:
            return;
        case NILE_TYPE_INDEXED:
            break;
        case NILE_TYPE_DIRECT:
            break;
        case NILE_TYPE_BITPLANES:
            break;
        case NILE_TYPE_BITPLANES_DIRECT:
            break;
        }
        break;
    }
}
